import re
from string import ascii_uppercase, digits
from typing import Any, Dict, Iterable, List, Set, Tuple, Union

from sign_language_translator.config.assets import Assets
from sign_language_translator.config.enums import TextLanguages
from sign_language_translator.languages.text.text_language import TextLanguage
from sign_language_translator.languages.vocab import Vocab
from sign_language_translator.text.preprocess import (
    remove_space_before_punctuation,
    replace_words,
)
from sign_language_translator.text.tagger import Rule, Tagger, Tags
from sign_language_translator.text.tokenizer import SignTokenizer

__all__ = [
    "Urdu",
]


class Urdu(TextLanguage):
    """NLP class for Urdu text. Extends `slt.languages.text.TextLanguage` class.

    Urdu is an Indo-Aryan language spoken mostly in Pakistan. Urdu uses the Perso-Arabic script, which consists of 46 Alphabets, 10 Digits, 6 Punctuations & 6 Diacritics, and is written from right to left.
    See unicode details at: https://unicode.org/charts/PDF/U0600.pdf
    """

    @staticmethod
    def name() -> str:
        return TextLanguages.URDU.value

    @classmethod
    def token_regex(cls) -> str:
        return cls.NUMBER_REGEX + r"|" + cls.WORD_REGEX

    @classmethod
    def allowed_characters(cls) -> Set[str]:
        return cls.ALLOWED_CHARACTERS

    def __init__(self) -> None:
        # TODO: args to filter dataset
        self.vocab = Vocab(
            language=r"^ur$",
            country=r".+",
            organization=r".+",
            part_number=r"[0-9]+",
            data_root_dir=Assets.ROOT_DIR,
            arg_is_regex=True,
        )

        self.non_sentence_end_tokens = {
            # letters (A.B.C.) & spelled out letters (Ay, Bee, See)
            w.upper()
            for wc in self.vocab.supported_words_with_word_sense
            for w in [self.vocab.remove_word_sense(wc)]
            if (("double-handed-letter)" in wc) and (not w.isascii()))
            or (len(w) == 1 and w.isalpha())
        }

        self.tokenizer = SignTokenizer(
            word_regex=self.token_regex(),
            compound_words=(
                self.vocab.supported_words
                | self.vocab.supported_words_with_word_sense
                | set(self.vocab.words_to_numbers.keys())
                | set(self.vocab.person_names)
            ),  # TODO: | one-hundred twenty-three (\d[ \d]*): ["100", "23"] --> ["123"]
            end_of_sentence_tokens=self.END_OF_SENTENCE_MARKS,
            full_stops=self.FULL_STOPS,
            non_sentence_end_words=self.non_sentence_end_tokens,
            tokenized_word_sense_pattern=[self.WORD_REGEX, r"\(", [r"نام"], r"\)"],
        )

        # :TODO: {<unk>: id_}, def token_to_id, tokenize(..., as_input_ids = True),

        self.tagging_rules = [
            # e.g. " "
            Rule.from_pattern(r"^\s$", Tags.SPACE, 5),
            # e.g. "," "."
            Rule.from_pattern(
                r"^[" + "".join(map(re.escape, self.PUNCTUATION)) + r"]$",
                Tags.PUNCTUATION,
                5,
            ),
            # e.g. "word"
            Rule.from_pattern("^" + self.token_regex() + "$", Tags.WORD, 5),
            # e.g. COVID
            Rule.from_pattern(r"^[A-Z]{2,7}$", Tags.ACRONYM, 4),
            # e.g. 2002-02-20
            Rule.from_pattern(r"^\d{4}-\d{2}-\d{2}$", Tags.DATE, 4),
            # e.g. 09:30:25.333
            Rule.from_pattern(r"^\d+(?::\d+)?(?::\d+(?:\.\d+)?)$", Tags.TIME, 4),
            # e.g. John, Doe(name)
            Rule(
                lambda token: token in self.vocab.person_names
                or token.endswith("(نام)"),
                Tags.NAME,
                2,
            ),
            # e.g. Cow, airplane, 1
            Rule(
                lambda token: (
                    token.lower() in self.vocab.supported_words
                    or token.lower() in self.vocab.supported_words_with_word_sense
                ),
                Tags.SUPPORTED_WORD,
                3,
            ),
            # e.g. forty-five, 45
            Rule(
                lambda token: (
                    bool(re.match(r"^\d+(?:\.\d+)?$", token))
                    or token in self.vocab.words_to_numbers
                ),
                Tags.NUMBER,
                4,
            ),
            # e.g. "میں" -> ["میں(i)", "میں(in)"]
            Rule(
                lambda token: token in self.vocab.ambiguous_to_unambiguous,
                Tags.AMBIGUOUS,
                2,
            ),
        ]
        self.tagger = Tagger(
            rules=self.tagging_rules,
            default=Tags.DEFAULT,
        )

    def preprocess(self, text: str) -> str:
        # TODO: optimize (especially regex)
        text = self.character_normalize(text)

        # spell fix
        text = replace_words(
            text,
            word_map=self.vocab.misspelled_to_correct,  # :TODO: split joint words
            word_regex=self.token_regex(),
        )
        text = self.delete_unallowed_characters(text)
        text = re.sub(r"[۔\.][۔\. ]+[\.۔]", "۔۔۔", text)
        text = re.sub(r"[ \t]+", " ", text)
        text = remove_space_before_punctuation(text, self.PUNCTUATION)
        text = text.strip()

        return text

    def tokenize(self, text: str) -> List[str]:
        tokens = self.tokenizer.tokenize(
            text, join_compound_words=True, join_word_sense=True
        )
        return tokens

    def sentence_tokenize(self, text: str) -> List[str]:
        sentences = self.tokenizer.sentence_tokenize(text)
        if len(sentences) > 1:
            sentences[1:] = [sentence.lstrip() for sentence in sentences[1:]]
            sentences[:-1] = [sentence.rstrip() for sentence in sentences[:-1]]
        # sentences = [sen for sen in sentences if sen]
        return sentences

    def detokenize(self, tokens: Iterable[str]) -> str:
        text = self.tokenizer.detokenize(tokens)
        return text

    def tag(self, tokens: Union[str, Iterable[str]]) -> List[Any]:
        if isinstance(tokens, str):
            tokens = [tokens]

        tagged = self.tagger.tag(tokens)

        return tagged

    def get_tags(self, tokens: Union[str, Iterable[str]]) -> List[Any]:
        if isinstance(tokens, str):
            tokens = [tokens]

        tags = self.tagger.get_tags(tokens)

        return tags

    def get_word_senses(self, tokens: Union[str, Iterable[str]]) -> List[List[str]]:
        if isinstance(tokens, str):
            tokens = [tokens]

        word_senses = [
            self.vocab.ambiguous_to_unambiguous.get(token, []) for token in tokens
        ]

        return word_senses

    # ====================== #
    #    Character Groups    #
    # ====================== #

    UNICODE_RANGE: Tuple[int, int] = (1536, 1791)

    FULL_STOPS: List[str] = [".", "۔"]
    QUESTION_MARKS: List[str] = ["?", "؟"]
    END_OF_SENTENCE_MARKS: List[str] = FULL_STOPS + QUESTION_MARKS + ["!"]

    PUNCTUATION: List[str] = END_OF_SENTENCE_MARKS + [",", "٫", "،", "؛"]

    QUOTATION_MARKS = """ ' " ” “ ’ ‘ """.split()
    BRACKETS: List[str] = ["(", ")"]
    SYMBOLS: List[str] = PUNCTUATION + QUOTATION_MARKS + BRACKETS + [" ", "-"]

    PUNCTUATION_REGEX = r"[" + "".join([re.escape(punc) for punc in PUNCTUATION]) + r"]"
    DIACRITICS = str(" ٍ ً ٰ َ ُ ِ ").split()
    EXTRA_DIACRITICS = str(" ؐ  ؑ ؒ ؓ").split()
    WORD_REGEX = r"[\w" + "".join(DIACRITICS) + r"]+"
    # TODO: r"[[^\W\d_]"+ "".join(DIACRITICS) + r"]+"

    NUMBER_REGEX = r"\d+(?:[\.:]\d+)*"

    CHARACTER_TO_WORD = {
        "ﷲ": "اللہ",
        "ﷺ": "صلی اللہ علیہ وسلم",
        "﷽": "بسم اللہ الرحمن الرحیم",
        "–": "-",
        "—": "-",
        "−": "-",
        "⋯": "...",
    }

    def delete_unallowed_characters(self, text: str) -> str:
        text = re.sub(self.UNALLOWED_CHARACTERS_REGEX, " ", text)

        return text

    # functions to preprocess specific datasets
    @staticmethod
    def poetry_preprocessor(text: str) -> str:
        text = ("؛ ").join(
            [
                line.strip("() '\"\t")
                for line in text.splitlines()
                if len(re.findall(Urdu.WORD_REGEX, line)) > 1
            ]
        )
        MISSPELLED_TO_CORRECT = {
            "مرا": "میرا",
            "مری": "میری",
            "مري": "میری",
            "مرے": "میرے",
        }
        text = replace_words(
            text,
            word_map=MISSPELLED_TO_CORRECT,
            word_regex=Urdu.WORD_REGEX,
        )

        return text

    @staticmethod
    def passage_preprocessor(text: str) -> str:
        text = re.sub(r"\s+", " ", text)
        text = text.strip()

        return text

    @staticmethod
    def wikipedia_preprocessor(text: str) -> str:
        text = text.strip(". !\"'\n\t")
        return text

    if True or "UrduHack":
        # Start of the code borrowed from "UrduHack/normalization/character.py"
        # """Following Dictionaries and code are copied from UrduHack package
        # Source Repo URL: https://github.com/urduhack/urduhack
        # Source Repo URL: https://github.com/urduhack/urdu-characters"""

        # Maps correct Urdu characters to list of visually similar non-urdu characters
        CORRECT_URDU_CHARACTERS_TO_INCORRECT: Dict[str, List[str]] = {
            "آ": ["ﺁ", "ﺂ"],
            "أ": ["ﺃ"],
            "ا": ["ﺍ", "ﺎ"],
            "ب": ["ﺏ", "ﺐ", "ﺑ", "ﺒ"],
            "پ": ["ﭖ", "ﭘ", "ﭙ"],
            "ت": ["ﺕ", "ﺖ", "ﺗ", "ﺘ"],
            "ٹ": ["ﭦ", "ﭧ", "ﭨ", "ﭩ"],
            "ث": ["ﺛ", "ﺜ", "ﺚ"],
            "ج": ["ﺝ", "ﺞ", "ﺟ", "ﺠ"],
            "ح": ["ﺡ", "ﺣ", "ﺤ", "ﺢ"],
            "خ": ["ﺧ", "ﺨ", "ﺦ"],
            "د": ["ﺩ", "ﺪ"],
            "ذ": ["ﺬ", "ﺫ"],
            "ر": ["ﺭ", "ﺮ"],
            "ز": ["ﺯ", "ﺰ"],
            "س": ["ﺱ", "ﺲ", "ﺳ", "ﺴ"],
            "ش": ["ﺵ", "ﺶ", "ﺷ", "ﺸ"],
            "ص": ["ﺹ", "ﺺ", "ﺻ", "ﺼ"],
            "ض": ["ﺽ", "ﺾ", "ﺿ", "ﻀ"],
            "ط": ["ﻃ", "ﻄ"],
            "ظ": ["ﻅ", "ﻇ", "ﻈ"],
            "ع": ["ﻉ", "ﻊ", "ﻋ", "ﻌ"],
            "غ": ["ﻍ", "ﻏ", "ﻐ"],
            "ف": ["ﻑ", "ﻒ", "ﻓ", "ﻔ"],
            "ق": ["ﻕ", "ﻖ", "ﻗ", "ﻘ"],
            "ل": ["ﻝ", "ﻞ", "ﻟ", "ﻠ"],
            "م": ["ﻡ", "ﻢ", "ﻣ", "ﻤ"],
            "ن": ["ﻥ", "ﻦ", "ﻧ", "ﻨ"],
            "چ": ["ﭺ", "ﭻ", "ﭼ", "ﭽ"],
            "ڈ": ["ﮈ", "ﮉ"],
            "ڑ": ["ﮍ", "ﮌ"],
            "ژ": ["ﮋ"],
            "ک": ["ﮎ", "ﮏ", "ﮐ", "ﮑ", "ﻛ", "ك"],
            "گ": ["ﮒ", "ﮓ", "ﮔ", "ﮕ"],
            "ں": ["ﮞ", "ﮟ"],
            "و": ["ﻮ", "ﻭ", "ﻮ"],
            "ؤ": ["ﺅ"],
            "ھ": ["ﮪ", "ﮬ", "ﮭ", "ﻬ", "ﻫ", "ﮫ"],
            "ہ": ["ﻩ", "ﮦ", "ﻪ", "ﮧ", "ﮩ", "ﮨ", "ه"],
            "ۂ": [],
            "ۃ": ["ة"],
            "ء": ["ﺀ"],
            "ی": ["ﯼ", "ى", "ﯽ", "ﻰ", "ﻱ", "ﻲ", "ﯾ", "ﯿ", "ي"],
            "ئ": ["ﺋ", "ﺌ"],
            "ے": ["ﮮ", "ﮯ", "ﻳ", "ﻴ"],
            "ۓ": [],
            "۰": ["٠"],
            "۱": ["١"],
            "۲": ["٢"],
            "۳": ["٣"],
            "۴": ["٤"],
            "۵": ["٥"],
            "۶": ["٦"],
            "۷": ["٧"],
            "۸": ["٨"],
            "۹": ["٩"],
            "۔": [],
            "؟": [],
            "٫": [],
            "،": [],
            "لا": ["ﻻ", "ﻼ"],
            # "": ["ـ"],
        }

        # Maps (character + diacritic) to single characters (beware RTL text rendering)
        SPLIT_TO_COMBINED_CHARACTERS: Dict[str, str] = {
            "آ": "آ",
            "أ": "أ",
            "ؤ": "ؤ",
            "ۂ": "ۂ",
            "یٔ": "ئ",
            "ۓ": "ۓ",
        }

        # Convert the dictionaries to a useable format
        CHARACTER_TRANSLATOR = {ord(c): w for c, w in CHARACTER_TO_WORD.items()} | {
            ord(non_urdu): urdu
            for urdu, others in CORRECT_URDU_CHARACTERS_TO_INCORRECT.items()
            for non_urdu in others
        }
        COMBINE_CHARACTERS_REGEX = r"|".join(SPLIT_TO_COMBINED_CHARACTERS.keys())
        DIACRITICS_REGEX = r"|".join(DIACRITICS)

        @staticmethod
        def character_normalize(text: str) -> str:
            """Replace characters that are rendered the same as Urdu characters in common fonts but actually belong to foreign unicode character ranges by Urdu characters.

            Args:
                text (str): a piece of urdu text that may contain foreign symbols

            Returns:
                str: normalized urdu text
            """

            text = text.translate(Urdu.CHARACTER_TRANSLATOR)
            text = re.sub(
                Urdu.COMBINE_CHARACTERS_REGEX,
                lambda match: Urdu.SPLIT_TO_COMBINED_CHARACTERS[match.group()],
                text,
            )

            return text

        @staticmethod
        def remove_diacritics(text: str) -> str:
            text = re.sub(Urdu.DIACRITICS_REGEX, "", text)

            return text

        # End of the code borrowed from UrduHack

    ALLOWED_CHARACTERS = (
        set(CORRECT_URDU_CHARACTERS_TO_INCORRECT.keys())
        | set(DIACRITICS)
        | set(SYMBOLS)
        | set(ascii_uppercase)  # acronyms
        | set(digits)
        | set(EXTRA_DIACRITICS)
        | set("()!.,?/[]{} \n")
    )
    UNALLOWED_CHARACTERS_REGEX = (
        "[^" + "".join(map(re.escape, ALLOWED_CHARACTERS)) + "]"
    )



"""This module provides BeamSampling class for generating completions using beam search sampling."""

from math import exp, log2
from random import random
from typing import Any, Callable, Iterable, Tuple

from sign_language_translator.models.language_models.abstract_language_model import (
    LanguageModel,
)
from sign_language_translator.utils import sample_one_index


class BeamSampling:
    """BeamSampling class for generating completions using beam search sampling.

    Args:
        model (LanguageModel): The language model used for generating completions.
        beam_width (int, optional): The beam width for beam search. Defaults to 3.
        start_of_sequence_token (str, optional): The start of sequence token. Defaults to "[".
        end_of_sequence_token (str, optional): The end of sequence token. Defaults to "]".
        max_length (int, optional): The maximum length of the generated completions. Defaults to 33.
        scoring_function (Callable[[Iterable, float], float], optional):
            The scoring function used to score the completions. It should accept the generated
            sequence and its overall log probability as arguments. Defaults to a linear function.

        return_log_of_probability (bool, optional):
            A flag indicating whether to return the probability of the completions
            or log2 of it. Defaults to True.
    """

    def __init__(
        self,
        model: LanguageModel,
        beam_width: int = 3,
        start_of_sequence_token="[",
        end_of_sequence_token="]",
        max_length: int = 37,
        scoring_function: Callable[[Iterable, float], float] = (
            lambda seq, log_prob: 10.0 + log_prob / len(seq)  # type: ignore
        ),
        return_log_of_probability: bool = True,
    ) -> None:
        self.model = model
        self.start_of_sequence_token = start_of_sequence_token
        self.end_of_sequence_token = end_of_sequence_token
        self.beam_width = beam_width
        self.max_length = max_length
        self.scoring_function = scoring_function
        self.return_log_of_probability = return_log_of_probability

    def __call__(self, context: Iterable | None = None) -> Iterable:
        return self.complete(initial_context=context)

    def complete(
        self,
        initial_context: Iterable | None = None,
        append_func: Callable[[Any, Any], Any] = lambda context, token: (
            context + [token]
            if isinstance(context, list)
            else context + (token,)
            if isinstance(context, tuple)
            else context + token
        ),
    ) -> Tuple[Iterable, float]:
        """Generate completions based on the given initial context.

        Args:
            initial_context (Iterable | None, optional): The initial context for completion generation. Defaults to None.
            append_func (Callable[[Any, Any], Any], optional): a function that can append the generated next token to provided context.
                Defaults to a lambda function that can append to list, tuple & str.

        Returns:
            Tuple[Iterable, float]: One generated completion and its score.
        """

        if initial_context is None:
            initial_context = [self.start_of_sequence_token]

        branches = [(initial_context, 0.0)]

        for _ in range(self.max_length):
            n_branches = round(self.beam_width + random() * 0.8 - 0.4)

            # Expand
            new_branches = []
            for context_, score_ in branches:
                if (  # sequence completed
                    context_[-1] == self.end_of_sequence_token  # type: ignore
                    or len(context_) >= self.max_length  # type: ignore
                ):
                    # just as it was. no change.
                    new_branches.append((context_, score_))
                    continue

                # append next tokens
                for _ in range(n_branches):
                    next_token, prob = self.model.next(context_)
                    if next_token == self.model.unknown_token:
                        score = score_
                        next_context = context_
                    else:
                        next_context = append_func(context_, next_token)
                        score = score_ + log2(prob)

                    # no repeats; for diversity.
                    if (next_context, score) not in new_branches:
                        new_branches.append((next_context, score))

            # Prune
            new_branches = sorted(
                new_branches,
                key=lambda item: self.scoring_function(*item),
                reverse=True,
            )[:n_branches]

            if branches == new_branches:
                # no branch has grown further: stop
                break

            branches = new_branches

        # softmax: turn each branch's scores into a probability distribution
        weights = [self.scoring_function(seq, log_prob) for seq, log_prob in branches]
        weights = [exp(w) for w in weights]
        weights = [w / sum(weights) for w in weights]

        # select one branch
        selected_completion, score = branches[sample_one_index(weights)]

        # reformat score
        if not self.return_log_of_probability:
            score = 2 ** score # math.exp2(score)

        return selected_completion, score



from abc import ABC, abstractmethod
from typing import Any, Iterable, List, Set, Tuple, Union


class TextLanguage(ABC):
    """Base NLP class for a language.

    Subclass it and provide the functionality to tokenize text and classify & disambiguate tokens.
    Each token should correspond to a sign language clip.
    """

    @staticmethod
    @abstractmethod
    def name() -> str:
        """Returns the name of the language used everywhere else in datasets."""

    @classmethod
    @abstractmethod
    def token_regex(cls) -> str:
        """Returns a regular expression that matches words in this language."""

    @classmethod
    @abstractmethod
    def allowed_characters(cls) -> Set[str]:
        """Returns a set of all allowed characters in the language."""

    @abstractmethod
    def preprocess(self, text: str) -> str:
        """Preprocesses text before tokenization.
        Make sure no different unicode characters are used for the same word.
        Remove unnecessary symbols, spaces, etc."""

    @abstractmethod
    def tokenize(self, text: str) -> List[str]:
        """Break apart text into words or phrases"""

    @abstractmethod
    def sentence_tokenize(self, text: str) -> List[str]:
        """Break text into sentences."""

    @abstractmethod
    def detokenize(self, tokens: Iterable[str]) -> str:
        """Joins tokens back into text."""

    @abstractmethod
    def tag(self, tokens: Union[str, Iterable[str]]) -> List[Tuple[str, Any]]:
        """Classify the tokens and mark them with appropriate tags."""

    @abstractmethod
    def get_tags(self, tokens: Union[str, Iterable[str]]) -> List[Any]:
        """Get the classifications of all tokens in the form of a sequence of tags"""

    @abstractmethod
    def get_word_senses(self, tokens: Union[str, Iterable[str]]) -> List[List[str]]:
        """Get all known meanings of the ambiguous words."""

    # embed/similar
    # all_tags



"""This module provides access to configuration components for the Sign Language Translator package.

Contents:
- Assets: Class to manage the assets of the Sign Language Translator package.
- enums: Defines enumerations of short codes used in the Sign Language Translator package.
- settings: Contains constants and configurations for the Sign Language Translator package.
- utils: Contains helper functions for configuration of the package.
"""

from sign_language_translator.config import enums, utils
from sign_language_translator.config.assets import Assets
from sign_language_translator.config.settings import Settings

__all__ = [
    # classes
    "Assets",
    "Settings",
    # modules
    "enums",
    "utils",
]



from sign_language_translator.vision import utils
from sign_language_translator.vision._utils import get_sign_wrapper_class
from sign_language_translator.vision.sign.sign import Sign
from sign_language_translator.vision.video.video import Video

__all__ = [
    # modules
    "utils",
    # classes
    "Sign",
    "Video",
    # functions
    "get_sign_wrapper_class",
]



"""utility to print out directory hierarchy

Bugs:
    multiple extra_line printed when parameter is True in nested calls
"""

import os
import re
from glob import glob


def tree(
    cur_path: str = ".",
    directory_only=True,
    extra_line=True,
    ignore=["__pycache__", "temp", "__init__.py"],
    regex=True,
) -> None:
    """prints out directory hierarchy

    Args:
        cur_path (str, optional): the root node of tree or the starting parent directory. Defaults to ".".
        directory_only (bool, optional): True means files will not be listed, only folders. Defaults to True.
    """

    # TODO: ANSI colors: branches, directories, files

    def make_tree(
        cur_path: str,
        prev_indent: str,
        directory_only: bool,
        extra_line=True,
        ignore=None,
        regex=True,
    ) -> None:
        """actually makes the directory hierarchy

        Args:
            cur_path (str): the root node of tree or the starting parent directory.
            prev_indent (str): Whatever string was printed behind the provided directory's name
            directory_only (bool): True means files will not be listed, only folders.
        """
        # TODO: glob patterns

        if ignore is None:
            ignore = ["__pycache__", "temp"]

        # list and sort subdirectories & contained files
        children_paths = glob(os.path.join(cur_path, "*")) # TODO: use os.listdir to show hidden files
        child_files = []
        child_dirs = []
        for child in children_paths:
            filename = os.path.split(child)[-1]
            if (regex and any(re.match(regex, filename) for regex in ignore)) or (
                (not regex) and (filename in ignore)
            ):
                continue

            if os.path.isdir(child):
                child_dirs.append(filename)
            else:
                if not directory_only:
                    child_files.append(filename)

        total_children = (
            len(child_dirs) if directory_only else len(child_dirs) + len(child_files)
        )

        for i, child in enumerate(sorted(child_files) + sorted(child_dirs)):
            if i < total_children - 1:
                child_indent, grandchild_indent = "├── ", "│   "
            else:
                child_indent, grandchild_indent = "└── ", "    "

            print(prev_indent + child_indent + child)
            make_tree(
                os.path.join(cur_path, child),
                prev_indent + grandchild_indent,
                directory_only,
                extra_line=extra_line,
                ignore=ignore,
            )

        if extra_line and (total_children > 0):
            line = prev_indent.rstrip()
            if line != "":
                print(line)

    print(os.path.split(os.path.abspath(cur_path))[-1])
    make_tree(
        os.path.abspath(cur_path),
        "",
        directory_only,
        extra_line=extra_line,
        ignore=ignore,
        regex=regex,
    )



"""This module provides an abstract base class for language models,
which can be used to sample the next token based on a given context.
It defines the common interface and methods that should be implemented by any language model.

Classes:
    LanguageModel: An abstract base class for language models.
"""

from abc import ABC, abstractmethod
from typing import Any, Iterable, Tuple


class LanguageModel(ABC):
    """Abstract Base Class for Language Models

    LanguageModel is an abstract base class that defines the common interface and methods for language models. It provides functionality for sampling the next token based on the given context.

    Attributes:
    - unknown_token (str): The token representation used for unknown or out-of-vocabulary tokens.
    - name (str): The name of the language model (optional).

    Methods:
    - next(self, context: Iterable) -> Tuple[Any, float]: Abstract method that should be implemented
        by subclasses to generate the next token and provide its probability.
    - next_all(self, context: Iterable) -> Tuple[Iterable[Any], Iterable[float]]: Abstract method
        that should be implemented by subclasses to return all next tokens and their probabilities.
    """

    def __init__(self, unknown_token="<unk>", name=None) -> None:
        super().__init__()
        self.unknown_token = unknown_token
        self.name = name

    @abstractmethod
    def next(self, context: Iterable) -> Tuple[Any, float]:
        """Generates the next token based on the given context and also returns its probability.

        Args:
            context (Iterable): A piece of sequence used as the context
                for generating the next token.

        Returns:
            Tuple[Any, float]: The next token and its associated probability.
                Token has the same type as the items in the context iterable.
        """

    @abstractmethod
    def next_all(self, context: Iterable) -> Tuple[Iterable[Any], Iterable[float]]:
        """Computes probabilities for all next tokens based on the given context and returns them both.

        Args:
            context (Iterable): A piece of sequence used as the context
                for generating the next tokens.

        Returns:
            Tuple[Iterable[Any], Iterable[float]]: All next tokens and their probabilities.
                The tokens have the same type as the items in the context iterable.
        """

    def __str__(self) -> str:
        return (
            f'name="{self.name}", ' if self.name else ""
        ) + f'unk_tok="{self.unknown_token}"'

    def __repr__(self) -> str:
        return str(self.__class__).split()[1][:-1] + "\n\n" + self.__str__()



"""Module that contains Text Language Processors as classes to clean up, tokenize and tag texts of various languages."""

from sign_language_translator.languages.text.hindi import Hindi
from sign_language_translator.languages.text.text_language import TextLanguage
from sign_language_translator.languages.text.urdu import Urdu
from sign_language_translator.text.tagger import Tags

__all__ = [
    "TextLanguage",
    "Urdu",
    "Hindi",
    "Tags",
]



from typing import Type

from sign_language_translator.config.enums import SignFormats, normalize_short_code
from sign_language_translator.vision.sign.sign import Sign

__all__ = [
    "get_sign_wrapper_class",
]


def get_sign_wrapper_class(sign_code: str) -> Type[Sign]:
    """
    Retrieves a Sign wrapping class based on the provided string code.

    Args:
        sign_code (str): The name of the Sign wrapper class. e.g. "video", "landmarks" etc.

    Returns:
        Type[Sign]: An instance of the Sign wrapper class corresponding to the provided code.

    Raises:
        ValueError: If no Sign wrapper class is known for the provided code.
    """

    from sign_language_translator.vision.video.video import Video

    code_to_class = {
        SignFormats.VIDEO.value: Video,
    }

    class_ = code_to_class.get(normalize_short_code(sign_code), None)
    if class_ is not None:
        return class_

    # Unknown
    raise ValueError(f"no sign wrapper class known for '{sign_code = }'")



from sign_language_translator.models.text_to_sign.concatenative_synthesis import (
    ConcatenativeSynthesis,
)
from sign_language_translator.models.text_to_sign.t2s_model import TextToSignModel

__all__ = [
    "TextToSignModel",
    "ConcatenativeSynthesis",
]



import re
from string import ascii_uppercase, digits
from typing import Any, Dict, Iterable, List, Set, Tuple, Union

from sign_language_translator.config.assets import Assets
from sign_language_translator.config.enums import TextLanguages
from sign_language_translator.languages.text.text_language import TextLanguage
from sign_language_translator.languages.vocab import Vocab
from sign_language_translator.text.preprocess import remove_space_before_punctuation
from sign_language_translator.text.tagger import Rule, Tagger, Tags
from sign_language_translator.text.tokenizer import SignTokenizer


class Hindi(TextLanguage):
    """NLP class for Hindi text. Extends `slt.languages.text.TextLanguage` class.

    Hindi is an Indo-Aryan language spoken mostly in India. Hindi uses the Devanagari script, which consists of 11 vowels and 33 consonants and is written from left to right.
    See unicode details at: https://unicode.org/charts/PDF/U0900.pdf
    """

    @staticmethod
    def name() -> str:
        return TextLanguages.HINDI.value

    @classmethod
    def token_regex(cls) -> str:
        return f"({cls.NUMBER_REGEX}|{cls.WORD_REGEX})"

    @classmethod
    def allowed_characters(cls) -> Set[str]:
        return cls.ALLOWED_CHARACTERS

    def __init__(self) -> None:
        # TODO: args to filter dataset
        self.vocab = self.__get_vocab()
        self.tokenizer = self.__get_tokenizer()
        self.tagging_rules = self.__get_tagging_rules()
        self.tagger = Tagger(rules=self.tagging_rules, default=Tags.DEFAULT)
        self.omitted_tokens = {"", " ", "\t"}

    def preprocess(self, text: str) -> str:
        text = self.normalize_characters(text)
        text = self.delete_unallowed_characters(text)
        text = re.sub(r"[^\S\n]{2,}", " ", text)
        text = remove_space_before_punctuation(text, self.PUNCTUATION)
        text = text.strip()

        return text

    def tokenize(self, text: str) -> List[str]:
        tokens = self.tokenizer.tokenize(
            text, join_compound_words=True, join_word_sense=True
        )
        tokens = [token for token in tokens if token not in self.omitted_tokens]
        return tokens

    def sentence_tokenize(self, text: str) -> List[str]:
        sentences = self.tokenizer.sentence_tokenize(text)
        sentences = [
            sentence.strip()
            for sentence in sentences
            if sentence not in self.omitted_tokens
        ]
        return sentences

    def detokenize(self, tokens: Iterable[str]) -> str:
        text = " ".join(tokens)
        text = remove_space_before_punctuation(text, self.PUNCTUATION)

        return text

    def tag(self, tokens: Union[str, Iterable[str]]) -> List[Tuple[str, Any]]:
        if isinstance(tokens, str):
            tokens = [tokens]

        tagged = self.tagger.tag(tokens)

        return tagged

    def get_tags(self, tokens: Union[str, Iterable[str]]) -> List[Any]:
        if isinstance(tokens, str):
            tokens = [tokens]

        tags = self.tagger.get_tags(tokens)

        return tags

    def get_word_senses(self, tokens: Union[str, Iterable[str]]) -> List[List[str]]:
        if isinstance(tokens, str):
            tokens = [tokens]

        word_senses = [
            self.vocab.ambiguous_to_unambiguous.get(token, []) for token in tokens
        ]

        return word_senses

    # ================== #
    #     Characters     #
    # ================== #

    UNICODE_RANGE: Tuple[int, int] = (2304, 2431)

    FULL_STOPS: List[str] = [".", "।", "॥"]
    QUESTION_MARKS: List[str] = ["?"]
    ACRONYM_PERIODS: List[str] = ["॰"]
    END_OF_SENTENCE_MARKS: List[str] = FULL_STOPS + QUESTION_MARKS + ["!"]

    PUNCTUATION: List[str] = END_OF_SENTENCE_MARKS + list("॰-_,!()[]/{}")

    CHARACTERS: List[str] = str(
        """
        ऀ  ँ  ं  ः  ऄ  अ  आ  इ  ई  उ  ऊ  ऋ  ऌ  ऍ  ऎ  ए
        ऐ  ऑ  ऒ  ओ  औ  क  ख  ग  घ  ङ  च  छ  ज  झ  ञ  ट
        ठ  ड  ढ  ण  त  थ  द  ध  न  ऩ  प  फ  ब  भ  म  य
        र  ऱ  ल  ळ  ऴ  व  श  ष  स  ह  ऺ  ऻ  ़  ऽ  ा  ि
        ी  ु  ू  ृ  ॄ  ॅ  ॆ  े  ै  ॉ  ॊ  ो  ौ  ्  ॎ  ॏ
        ॐ  ॑  ॒  ॓  ॔  ॕ  ॖ  ॗ  क़  ख़  ग़  ज़  ड़  ढ़  फ़  य़
        ॠ  ॡ  ॢ  ॣ  ।  ॥  ०  १  २  ३  ४  ५  ६  ७  ८  ९
        ॰  ॱ  ॲ  ॳ  ॴ  ॵ  ॶ  ॷ  ॸ  ॹ  ॺ  ॻ  ॼ  ॽ  ॾ  ॿ
        """
    ).split()
    DIACRITICS = str("ऀ  ँ  ं  ः  ॄ  ॅ  ़  ा  ि  ी  ु  ू  ृ  े  ै  ॉ  ो  ौ  ्").split()

    ALLOWED_CHARACTERS: Set[str] = (
        set(CHARACTERS)
        | set(DIACRITICS)
        | set(PUNCTUATION)
        | set(ascii_uppercase)
        | set(digits)
        | set("()!.,?/[]{} \n")
    )

    CHARACTER_TO_DECOMPOSED: Dict[str, str] = {
        "क़": "क़",
        "ख़": "ख़",
        "ग़": "ग़",
        "ज़": "ज़",
        "ड़": "ड़",
        "ढ़": "ढ़",
        "फ़": "फ़",
        "य़": "य़",
    }
    CHARACTER_TRANSLATOR = {ord(c): d for c, d in CHARACTER_TO_DECOMPOSED.items()}

    # ========== #
    #    Regex   #
    # ========== #

    NUMBER_REGEX = r"(\d+(?:[\.:]\d+)*)"
    WORD_REGEX = r"([^\W_\d]([^\W_\d]|[" + "".join(DIACRITICS) + r"])*)"
    UNALLOWED_CHARACTERS_REGEX = (
        "[^" + "".join(map(re.escape, ALLOWED_CHARACTERS)) + "]"
    )

    # ====================== #
    #    Helper Functions    #
    # ====================== #

    def delete_unallowed_characters(self, text: str) -> str:
        text = re.sub(self.UNALLOWED_CHARACTERS_REGEX, " ", text)

        return text

    def normalize_characters(self, text: str) -> str:
        text = text.translate(self.CHARACTER_TRANSLATOR)
        return text

    # ================ #
    #    initialize    #
    # ================ #

    def __get_vocab(self):
        vocab = Vocab(
            language=f"^{self.name()}$",  # r"^hi$"
            country=r"[a-z]+",
            organization=r"[a-z]+",
            part_number=r"[0-9]+",
            data_root_dir=Assets.ROOT_DIR,
            arg_is_regex=True,
        )
        return vocab

    def __get_tokenizer(self):
        tokenizer = SignTokenizer(
            word_regex=self.token_regex(),
            compound_words=(
                self.vocab.supported_words
                | self.vocab.supported_words_with_word_sense
                | set(self.vocab.words_to_numbers.keys())
                | set(self.vocab.person_names)
            ),  # TODO: | one-hundred twenty-three (\d[ \d]*): ["100", "23"] --> ["123"]
            end_of_sentence_tokens=self.END_OF_SENTENCE_MARKS,
            full_stops=self.FULL_STOPS,
            non_sentence_end_words=[
                "बी",
                "सी",
                "एस",
            ],  # spelled out english letters (acronyms)
            tokenized_word_sense_pattern=[self.WORD_REGEX, r"\(", [r"नाम"], r"\)"],
        )
        return tokenizer

    def __get_tagging_rules(self):
        punctuation_set = set(self.PUNCTUATION)
        tagging_rules = [
            # e.g. " "
            Rule.from_pattern(r"^\s+$", Tags.SPACE, 5),
            # e.g. "," "."
            Rule(lambda token: token in punctuation_set, Tags.PUNCTUATION, 5),
            # e.g. "word"
            Rule.from_pattern("^" + self.WORD_REGEX + "$", Tags.WORD, 5),
            # e.g. COVID
            Rule.from_pattern(r"^[A-Z]{2,8}$", Tags.ACRONYM, 4),
            # e.g. 2002-02-20
            Rule.from_pattern(r"^\d{4}-\d{2}-\d{2}$", Tags.DATE, 4),
            # e.g. 09:30:25.333
            Rule.from_pattern(r"^\d+(?::\d+)?(?::\d+(?:\.\d+)?)$", Tags.TIME, 4),
            # e.g. John, Doe(name)
            Rule(
                lambda token: token in self.vocab.person_names
                or token.endswith("(नाम)"),
                Tags.NAME,
                2,
            ),
            # e.g. Cow, airplane, 1
            Rule(
                lambda token: (
                    token.lower() in self.vocab.supported_words
                    or token.lower() in self.vocab.supported_words_with_word_sense
                ),
                Tags.SUPPORTED_WORD,
                3,
            ),
            # e.g. forty-five, 45
            Rule(
                lambda token: (
                    bool(re.match(r"^\d+(?:\.\d+)?$", token))
                    or token in self.vocab.words_to_numbers
                ),
                Tags.NUMBER,
                4,
            ),
            # e.g. "सोना" -> ["सोना(gold)", "सोना(sleep)"]
            Rule(
                lambda token: token in self.vocab.ambiguous_to_unambiguous,
                Tags.AMBIGUOUS,
                2,
            ),
        ]
        return tagging_rules



"""This module contains various helper functions and utilities for configuration purposes.

Module Structure:
- get_package_version() -> str: Retrieves the version of the package.
- read_urls(file_path: str) -> Dict[str, str]:
    reads urls for package resources and prepares filename to url map.
"""

import json
from importlib.metadata import version
from typing import Dict

from sign_language_translator.utils.utils import extract_recursive


def get_package_version():
    """
    Retrieves the version of the 'sign-language-translator' package.

    Returns:
        str: The version of the package.
    """

    return version("sign_language_translator")


def read_urls(file_path: str, encoding="utf-8") -> Dict[str, str]:
    # refactor: extract key recursively
    """
    Prepares a dictionary mapping filenames to their corresponding URLs.

    Args:
        file_path (str): The path to the JSON file containing resources information.

    Returns:
        Dict[str, str]: A dictionary mapping filenames to their corresponding URLs.
    """

    with open(file_path, "r", encoding=encoding) as f:
        data = json.load(f)

    filename_url_dict = {
        file: url
        for file_to_url in extract_recursive(data, "file_to_url")
        for file, url in file_to_url.items()
    }

    return filename_url_dict



from sign_language_translator.languages.utils import (
    get_text_language,
    get_sign_language,
)
from sign_language_translator.languages import text, sign

from sign_language_translator.languages.sign import SignLanguage
from sign_language_translator.languages.text import TextLanguage
from sign_language_translator.languages.vocab import Vocab

__all__ = [
    "get_text_language",
    "get_sign_language",
    "text",
    "sign",
    "TextLanguage",
    "SignLanguage",
    "Vocab",
]



"""This module provides the MixerLM class, which is a language model that allows sampling from multiple language models.

Classes:
- MixerLM: A language model that combines the outputs of multiple language models based on a selection strategy. It extends the abstract base class LanguageModel.
"""

from __future__ import annotations

import pickle
from os.path import exists
from typing import Any, Iterable, List, Tuple

from sign_language_translator.models.language_models.abstract_language_model import (
    LanguageModel,
)
from sign_language_translator.utils import sample_one_index


class MixerLM(LanguageModel):
    """The MixerLM class is a language model that combines multiple language models using a mixing strategy. It extends the abstract base class LanguageModel.

    Attributes:
    - models (List[LanguageModel]): List of language models to be combined.
    - selection_probabilities (List[float] | None): The selection probabilities for each language model.
                                If not provided, equal probabilities are assigned.
    - unknown_token (str): The token representation used for unknown or out-of-vocabulary tokens.
    - model_selection_strategy (str): The strategy for selecting the next token from the language models.
        Possible values: "choose" (selects one model and infers through it).
                         "merge"  (infers through all models & combines their output probabilities).
    - name (str): The name of the mixer language model object (optional).

    Methods:
    - next(self, context: Iterable) -> Tuple[Any, float]: Generates the next token based on the given context.
    - next_all(self, context: Iterable) -> Tuple[List[Any], List[float]]: Generates all next
        tokens and their associated probabilities based on the given context.
    - save(self, model_path: str) -> None: saves the mixer model as a pickle file.
    - load(model_path: str) -> MixerLM: loads the mixer model from a pickle file.
    - __str__(self) -> str: Returns a string representation of the MixerLM instance.
    """

    def __init__(
        self,
        models: List[LanguageModel],
        selection_probabilities: List[float] | None = None,
        unknown_token="<unk>",
        name=None,
        model_selection_strategy="choose" or "merge",
    ) -> None:
        if selection_probabilities is None:
            selection_probabilities = [1.0] * len(models)

        assert len(models) == len(selection_probabilities)
        assert all(isinstance(p, (float, int)) for p in selection_probabilities)

        super().__init__(unknown_token=unknown_token)
        self.models = models
        total = sum(selection_probabilities)
        self.selection_probabilities = [w / total for w in selection_probabilities]
        self.name = name
        assert model_selection_strategy in {"merge", "choose"}
        self.strategy = model_selection_strategy

    def next(self, context: Iterable) -> Tuple[Any, float]:
        token_and_prob = (self.unknown_token, 1.0)

        if self.strategy == "merge":
            tokens, probabilities = self.next_all(context)
            i = sample_one_index(probabilities)
            token_and_prob = tokens[i], probabilities[i]

        elif self.strategy == "choose":
            tried_models = set()
            while len(tried_models) < len(self.models):
                # select model
                model_idx = sample_one_index(self.selection_probabilities)
                if model_idx in tried_models:
                    continue
                tried_models.add(model_idx)

                # sample next token
                token, prob = self.models[model_idx].next(context)
                if token != self.models[model_idx].unknown_token:
                    token_and_prob = token, prob
                    break

        return token_and_prob

    def next_all(self, context: Iterable) -> Tuple[List[Any], List[float]]:
        """Computes probabilities for all next tokens based on the given context and returns them both.

        If model selection strategy is "choose" then selects one model and infers through it.
        If model_selection_strategy is "merge" then for each language model,
        it generates the all next tokens and probabilities. It combines the tokens and probabilities
        from all models to create a list of unique next tokens and their corresponding weighted probabilities.

        Args:
            context (Iterable): A piece of sequence like the training examples.

        Returns:
            Tuple[List[Any], List[float]]: A tuple containing a list of unique next tokens and their corresponding probabilities.
        """

        next_tokens = []
        probabilities = []
        total_weight = 1.0
        for model, weight in zip(self.models, self.selection_probabilities):
            all_token, all_probability = model.next_all(context)
            if model.unknown_token not in all_token:
                for token, prob in zip(all_token, all_probability):
                    if token not in next_tokens:
                        next_tokens.append(token)
                        probabilities.append(0)
                        i = -1
                    else:
                        i = next_tokens.index(token)
                    probabilities[i] += prob * weight
            else:
                total_weight -= weight

        if not next_tokens:
            return [self.unknown_token], [1.0]

        if total_weight != 1.0:
            probabilities = [p / total_weight for p in probabilities]
        assert (
            abs(sum(probabilities) - 1) < 0.001
        ), f"{probabilities= }\nsum = {sum(probabilities)}"

        return next_tokens, probabilities

    def save(self, model_path: str, overwrite=False) -> None:
        """
        Save the model to a file.

        Args:
            model_path (str): The path to save the model.
            overwrite (bool, optional): Whether to overwrite an existing file. Defaults to False.

        Raises:
            FileExistsError: If a file already exists at `model_path` and `overwrite` is False.
        """

        if exists(model_path) and not overwrite:
            raise FileExistsError(f"there is already a file at {model_path = }")

        with open(model_path, "wb") as f:
            pickle.dump(self, f)

    @staticmethod
    def load(model_path: str) -> MixerLM:
        """
        Loads a MixerLM model from the given model path.

        Parameters:
            model_path (str): The path to the model file.

        Returns:
            MixerLM: The loaded MixerLM model.
        """

        with open(model_path, "rb") as f:
            mixer_model = pickle.load(f)
        return mixer_model

    def __str__(self) -> str:
        # don't judge my code by its looks.. it works!
        model_strs = []
        for i, model in enumerate(self.models):
            branch = "├── " if i < len(self.models) - 1 else "└── "
            model_lines = str(model).splitlines()
            for line_no, _ in enumerate(model_lines):
                model_lines[line_no] = (
                    (
                        branch
                        if line_no == 0
                        else ("│   " if branch == "├── " else "    ")
                    )
                    + model_lines[line_no]
                    + (
                        f" | prob={self.selection_probabilities[i]:.1%}"
                        if line_no == 0
                        else ""
                    )
                )
            model_strs.append("\n".join(model_lines))

        mixer_str = f"Mixer LM: {super().__str__()}[{len(self.models)}]\n" + "\n".join(
            model_strs
        )
        return mixer_str



"""
Sign Language Translator (SLT) Command Line Interface

This module provides a command line interface (CLI) for the Sign Language Translator (SLT) library.
It allows you to perform various operations such as translating text to sign language or vice versa,
downloading resource files, completing text sequences using Language Models & embedding videos into
sequences of vectors.

.. code-block:: console

    $ slt
    Usage:
        slt [OPTIONS] COMMAND [ARGS]...

    Options:
        --help  Show this message and exit.

    Commands:
        assets     Assets manager to download & display Datasets & Models.
        complete   Complete a sequence using Language Models.
        translate  Translate text into sign language or vice versa.
        embed      Embed Videos Using Selected Model.
"""

import os

import click

from sign_language_translator import __version__

# TODO: Dockerize the CLI.. but model persistance issue


@click.group()
@click.version_option(__version__)
def slt():
    """
    Sign Language Translator (SLT) command line interface.
    Documentation: https://sign-language-translator.readthedocs.io
    """


@slt.group(no_args_is_help=True)
def assets():
    """
    Assets manager to download & display Datasets & Models.
    """


# Display the assets root directory
@assets.command()
def path():
    """
    Display the assets root directory.
    """

    from sign_language_translator.config.assets import Assets

    click.echo(Assets.ROOT_DIR)


# Display a tree of downloaded resource files
@assets.command()
@click.option(
    "--files",
    "-f",
    default=True,
    show_default=True,
    help="Include files as leaf nodes in the directory tree.",
)
@click.option(
    "--ignore",
    "-i",
    default=[],
    multiple=True,
    show_default=True,
    help="List of regular expressions of file and directory names that should not be displayed.",
)
@click.option(
    "--directory",
    "-d",
    default=None,
    help="Where the asset files are stored. Defaults to 'installation-directory/assets'.",
)
def tree(files, ignore, directory):
    """
    Display a hierarchy of files in the SLT Assets folder.

    Examples:\n
        $ slt assets tree\n
        $ slt assets tree -f false\n
        $ slt assets tree -i ".*\\.mp4" -i ".*\\.csv"
    """

    from sign_language_translator.config.assets import Assets
    from sign_language_translator.utils import tree as display_tree

    if directory:
        Assets.set_root_dir(directory)

    display_tree(Assets.ROOT_DIR, directory_only=not files, ignore=ignore, regex=True)


# Download resource files
@assets.command(no_args_is_help=True)
@click.argument("filenames", nargs=-1, required=True)
@click.option(
    "--overwrite",
    "-o",
    default=False,
    help="Overwrite existing files. Defaults to False.",
)
@click.option(
    "--progress-bar",
    "-p",
    default=True,
    help="Show progress bar for kilobytes downloaded. Defaults to True.",
)
@click.option(
    "--timeout",
    default=20,
    help="Timeout duration for download requests. Defaults to 20 sec.",
)
@click.option(
    "--chunk-size",
    default=131072,
    help="number of bytes to download in each step. Defaults to 131072 or 128K.",
)
@click.option(
    "--directory",
    "-d",
    default=None,
    help="Where to save the downloaded files. Defaults to 'installation-directory/assets'.",
)
def download(filenames, overwrite, progress_bar, timeout, chunk_size, directory):
    """
    Download resource files with regex.

    Downloads package resources matching the given filename regex
    and saves them to the resources folder in installation directory.\n
    Examples:\n
        $ slt download '.*.json' --overwrite true
    """

    from sign_language_translator.config.assets import Assets

    if directory:
        Assets.set_root_dir(directory)

    # Download the specified files
    for filename in filenames:
        success = Assets.download(
            filename,
            overwrite=overwrite,
            progress_bar=progress_bar,
            timeout=timeout,
            chunk_size=chunk_size,
        )
        if success:
            click.echo(f"Downloaded package resource(s) matching: '{filename}'")
        else:
            click.echo(f"Couldn't download package resources matching: '{filename}'")


# Translate
@slt.command(no_args_is_help=True)
@click.argument("inputs", nargs=-1, required=True)
@click.option(
    "--model-code",
    "-m",
    required=True,
    help="Short code to identify the translation model.",
)
@click.option("--text-lang", help="Name of the text language.")
@click.option("--sign-lang", help="Name of the sign language.")
@click.option(
    "--sign-format",
    help="the sign features to be used e.g. video, mediapipe-landmarks etc.",
)
@click.option(
    "--output-dir", default=".", help="Output directory for generated translations."
)
@click.option(
    "--overwrite",
    "-o",
    default=False,
    help="Whether to overwrite the target file if it already exists. Defaults to False.",
)
@click.option(
    "--display",
    "-d",
    default=True,
    help="Whether to show the output video. Defaults to True.",
)
@click.option(
    "--save-format",
    default="mp4",
    help="The output file extension. Defaults to 'mp4'.",
)
def translate(
    inputs,
    model_code,
    text_lang,
    sign_lang,
    sign_format,
    output_dir,
    overwrite,
    display,
    save_format,
):
    """
    Translate text into sign language or vice versa.

    Translates the given inputs between text and sign language using the specified model and features.
    Currently following model-codes are supported:\n
    1. "concatenative-synthesis" or "rule-based"\n
    Examples:\n
        $ slt translate --model-code rule-based --text-lang urdu --sign-lang psl --sign-format video "ایک سیب اچھا ہے۔"
    """

    from sign_language_translator import get_model, get_sign_wrapper_class
    from sign_language_translator.models import TextToSignModel

    model = get_model(
        model_code=model_code,
        sign_language=sign_lang,
        text_language=text_lang,
        sign_format=sign_format,
    )
    if model and isinstance(model, TextToSignModel):  # TODO: , SignToTextModel):
        for text in inputs:
            sign = model.translate(text)
            path = os.path.join(output_dir, f"{text}.{save_format}")
            sign.save(path, overwrite=overwrite, leave=True)
            if display:
                get_sign_wrapper_class(sign.name())(path).show()

    else:
        click.echo("This type of translation is not yet supported!")


# Complete
@slt.command(no_args_is_help=True)
@click.argument("inputs", nargs=-1, required=True)
@click.option(
    "--model-code",
    "-m",
    required=True,
    help="Short code to identify the language model. You can use multiple models (but of same input type).",
    multiple=True,
)
@click.option(
    "--end-token",
    default=">",
    help="Keep generating until this token.",
    show_default=True,
)
@click.option(
    "--max-length",
    default=20,
    help="Maximum number of tokens to generate.",
    show_default=True,
)
@click.option(
    "--beam-width",
    default=3.0,
    help="Number of possible branches to explore during generation.",
    show_default=True,
)
@click.option(
    "--model-weight",
    "-w",
    default=[1.0],
    multiple=True,
    help="Likelihood of this model getting selected in case multiple models are passed. Defaults to equal probabilities.",
)
@click.option(
    "--selection-strategy",
    default="choose",
    help="In case multiple models are used, should one model be selected at a time for inference ('choose') or should all models be inferred and their output probabilities combined before sampling ('merge').",
    show_default=True,
)
@click.option(
    "--join",
    default=None,
    help="Join the tokens by placing this string inbetween.",
)
def complete(
    inputs,
    model_code,
    end_token,
    max_length,
    beam_width,
    model_weight,
    selection_strategy,
    join,
):
    """
    Complete a sequence using Language Models.

    Predicts next tokens in the given sequence and completes it using the specified model and beam search.
    You can also use multiple models and also pass along model selection weights for each model (using --model-weights).
    Currently you can use the following model-codes:\n
    1. urdu-mixed-ngram (token | start: "<")\n
    2. ur-supported-gpt (token | start: "<")\n
    3. unigram-names (char | start: "[")\n
    4. bigram-names  (char | start: "[a")\n
    5. trigram-names (char | start: "[ab")\n
    Examples:\n
        $ slt complete "<" --model-code urdu-mixed-ngram -w 1.0 --model-code ur-supported-gpt -w 1.5
    """

    from sign_language_translator import get_model
    from sign_language_translator.models import BeamSampling, LanguageModel, MixerLM

    if model_weight == (1,):
        model_weight = [1.0] * len(model_code)
    assert len(model_code) == len(
        model_weight
    ), f"provide 1 weight for every model. Received {model_weight}. Size mismatch: {len(model_code)-len(model_weight)}."

    models = [get_model(model_code=code) for code in model_code]

    if all(isinstance(model, LanguageModel) for model in models):
        sampler = BeamSampling(
            MixerLM(
                models,  # type: ignore
                selection_probabilities=model_weight,
                model_selection_strategy=selection_strategy,
            ),
            beam_width=beam_width,
            end_of_sequence_token=end_token,
            max_length=max_length,
        )

        if models[0].name and "character" in models[0].name:  # type: ignore
            for inp in inputs:
                completion, _ = sampler.complete(inp)
                if isinstance(join, str):
                    completion = join.join(completion)
                click.echo(completion)
        else:
            # assume that all inputs are tokens of same sequence
            completion, _ = sampler.complete(inputs)
            if isinstance(join, str):
                completion = join.join(completion)
            click.echo(completion)

    else:
        click.echo("Model loading failed!")


# Embed
@slt.command(no_args_is_help=True)
@click.argument("inputs", nargs=-1, required=True)
@click.option(
    "--model-code",
    "-m",
    required=True,
    help="Short code to identify the embedding model.",
)
@click.option(
    "--processes",
    default=1,
    help="Number of processes to launch to embed using multiprocessing.",
)
@click.option(
    "--embedding-type",
    default="all",
    help="Optional parameter to filter down embeddings if model can return various types.",
)
@click.option(
    "--output-dir",
    default=".",
    help="where to save the generated embeddings.",
)
@click.option(
    "--save-format",
    default="pt",
    help="What format to save the generated embeddings in. (e.g. pt, csv, npy, npz). Defaults to 'pt'.",
)
@click.option(
    "--overwrite",
    default=False,
    help="Flag to indicate whether to overwrite existing files. Defaults to False.",
)
# TODO: Precision/format
# @click.option(
#     "--precision",
#     default=4,
#     help="Number of decimal places to save.",
#     show_default=True,
# )
def embed(
    inputs,
    model_code,
    processes: int,
    embedding_type,
    output_dir,
    save_format,
    overwrite,
):
    """
    Embed Videos Using Selected Model.

    This function processes input videos using a selected embedding model to generate video embeddings.
    It supports multiple input videos using path patterns and multiprocessing for efficiency.
    Currently you can use the following model-codes:\n
    1. mediapipe-pose-2-hand-1\n
    2. mediapipe-pose-1-hand-1\n
    3. mediapipe-pose-0-hand-1\n
    Example:\n
        $ slt embed dataset/*.mp4 --model-code mediapipe-pose-2-hand-1 --embedding-type all --processes 4 --save-format csv
    """

    from sign_language_translator import get_model
    from sign_language_translator.models.utils import VideoEmbeddingPipeline

    model = get_model(model_code)
    pipeline = VideoEmbeddingPipeline(model)  # type: ignore

    pipeline.process_videos_parallel(
        inputs,
        n_processes=processes,
        output_dir=output_dir,
        save_format=save_format,
        landmark_type=embedding_type,
        overwrite=overwrite,
    )


if __name__ == "__main__":
    slt()



"""
Module for downloading files from URLs and managing package resources.

This module provides functions for downloading files from specified URLs and saving them to the given file paths.
It also includes a function for downloading package resources matching a specified filename regex and saving them
to the appropriate file paths.

Functions:
- download(file_path, url, overwrite=False, progress_bar=False, timeout=20.0, chunk_size=65536):
    Downloads a file from the specified URL and saves it to the given file path.
- download_package_resource(filename_regex, overwrite=False, progress_bar=False, timeout=20.0):
    Downloads package resources matching the given filename regex and saves them to the appropriate file paths.
"""

import os
from time import time
from warnings import warn

import requests
from tqdm.auto import tqdm

__all__ = [
    "download",
]


def download(
    file_path: str,
    url: str,
    overwrite=False,
    progress_bar=False,
    timeout: float = 20.0,
    leave=True,
    chunk_size=65536,
    status_callback=None,
) -> bool:
    """
    Downloads a file from the specified URL and saves it to the given file path.

    Args:
        file_path (str): The path where the downloaded file will be saved.
        url (str): The URL of the file to be downloaded.
        overwrite (bool, optional): If False, skips downloading if the file already exists. Defaults to False.
        progress_bar (bool, optional): If True, displays a progress bar during the download. Defaults to False.
        timeout (int, optional): The maximum number of seconds to wait for a server response. Defaults to 20.0.
        leave (bool, optional): Wether to leave the progress bar behind after the download. Defaults to True.
        chunk_size (int, optional): The number of bytes to fetch in each step. Defaults to 65536.

    Returns:
        bool: True if the file is downloaded successfully, False otherwise.

    Raises:
        FileExistsError: if overwrite is False and the destination path already contains a file.
    """

    # TODO: resume failed download (headers = {'Range': f'bytes={resume_byte_pos}-'})
    # TODO: separate threads for download and writing (implement queue(s))

    if os.path.exists(file_path) and not overwrite:
        raise FileExistsError(f"There is already a file at {file_path}")

    try:
        response = requests.get(url, stream=True, timeout=timeout, allow_redirects=True)
        response.raise_for_status()

        if os.path.dirname(file_path):
            os.makedirs(os.path.dirname(file_path), exist_ok=True)

        with open(file_path, "wb") as fp:
            stream = response.iter_content(chunk_size=chunk_size)

            total_bytes = int(response.headers.get("content-length", 0))
            if progress_bar:
                stream = tqdm(
                    stream,
                    total=(total_bytes // chunk_size) + 1,
                    desc=f"Downloading {os.path.split(file_path)[-1]}",
                    leave=leave,
                    unit="chunk",
                )

            start_time = time()
            speed = 0
            saved = 0
            # download & write
            for chunk in stream:
                if chunk:
                    fp.write(chunk)

                # Calculate download speed (exponential moving average)
                speed += len(chunk) / (1024**2) / (time() - start_time)
                speed /= 2
                start_time = time()

                # Update progress bar
                if isinstance(stream, tqdm):
                    stream.set_postfix_str(f"{speed:.3f}MB/s")

                if status_callback:
                    saved += len(chunk)
                    status = {
                        "file": f"{saved / total_bytes:.1%}" if total_bytes else "?%",
                        "down": f"{speed:.3f}MB/s",
                    }
                    status_callback(status)

        # TODO: if hashlib.sha256(model_bytes).hexdigest() != expected_sha256:

        return True

    except requests.exceptions.RequestException as e:
        warn(f"Download Failed: {str(e)}")
        return False



from itertools import islice
from os.path import isfile
from time import sleep, time
from typing import Iterable
from warnings import warn

import cv2
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import Animation, ArtistAnimation
from numpy import uint8
from numpy.typing import NDArray

from sign_language_translator.utils import in_jupyter_notebook


class VideoDisplay:
    @staticmethod
    def display_ipython_video_in_jupyter(
        path: str,
        html_attributes="loop autoplay controls preload muted",
        width=640,
        height=480,
    ) -> None:
        if not in_jupyter_notebook():
            warn(f"Attempted to use IPython.core.display.Video outside jupyter: {path}")
            return

        from IPython.display import Video, display  # type: ignore

        if isfile(path):
            display(
                Video(path, width=width, height=height, html_attributes=html_attributes)
            )
        else:
            raise FileNotFoundError(f"No video file found at {path}")

    @staticmethod
    def frames_to_matplotlib_animation(
        frames: Iterable[NDArray[uint8]], fps: float = 30.0, close_plot=True
    ) -> Animation:
        fig, ax = plt.subplots()
        fig.set_facecolor("black")
        ax.axis("off")
        fig.tight_layout(pad=0)
        ims = []

        for frame in frames:
            im = ax.imshow(frame, animated=True)
            ims.append([im])

        animation = ArtistAnimation(fig, ims, interval=(1000 / fps), blit=True)

        if close_plot:
            plt.close()

        return animation

    @staticmethod
    def display_frames(
        frames: Iterable[NDArray[uint8]],
        fps: float = 30.0,
        inline_player="jshtml" or "html5",
        max_frames=None,
    ) -> None:
        if isinstance(frames, np.ndarray) and frames.ndim == 3:
            frames = [frames]

        if max_frames is not None:
            frames = islice(frames, max_frames)

        if in_jupyter_notebook():
            if (hasattr(frames, "__len__") and len(frames) == 1) or max_frames == 1:  # type: ignore
                VideoDisplay.display_frames_as_ipython_image(
                    [next(iter(frames))], fps=float("inf")
                )
            else:
                from IPython.display import HTML, display  # type: ignore

                animation = VideoDisplay.frames_to_matplotlib_animation(
                    frames, fps, close_plot=True
                )
                html = (
                    animation.to_html5_video()
                    if inline_player == "html5"
                    else animation.to_jshtml()
                )
                display(HTML(html))
        else:
            if (hasattr(frames, "__len__") and len(frames) == 1) or max_frames == 1:  # type: ignore
                VideoDisplay.show_image_plot(next(iter(frames)))  # type: ignore
            else:
                # TODO: Improve fps of the video player.
                # TODO: Add seek bar to the video player.
                animation = VideoDisplay.frames_to_matplotlib_animation(
                    frames, fps, close_plot=False
                )
            plt.show(block=False)  # TODO: window name
            plt.pause((len(frames) / fps) * 1.02 if hasattr(frames, "__len__") else 10)  # type: ignore
            plt.close()

    @staticmethod
    def display_frames_as_ipython_image(
        frames: Iterable[NDArray[uint8]], fps: float = 30.0
    ):
        # BUG: randomly prints None or shows Image

        from IPython.display import Image, display  # type: ignore

        display_handle = display(None, display_id=True)
        for frame in frames:
            start_time = time()
            # RGB to BGR because opencv is weird
            _, frame = cv2.imencode(".jpeg", cv2.cvtColor(frame, cv2.COLOR_BGR2RGB))
            display_handle.update(Image(data=frame.tobytes()))  # type: ignore

            elapsed_time = (start_time - time()) if start_time is not None else 0
            sleep(max(0, 1 / fps - elapsed_time))

    @staticmethod
    def show_image_plot(frame: NDArray[uint8]) -> None:
        plt.figure(facecolor="black")
        plt.axis("off")
        plt.tight_layout(pad=0)
        plt.imshow(frame)



"""parallel.py - Parallel Processing Utilities

This module provides utilities for performing parallel processing tasks using multithreading in Python.

Functions:
    - threaded_map: Multi-threaded mapping of a function to an iterable. Useful for I/O bound tasks.
"""

import threading
from time import sleep
from typing import Callable, Iterable

from tqdm.auto import tqdm


def threaded_map(
    target: Callable,
    args_list: Iterable,
    time_delay=0.02,
    timeout: float | None = None,
    max_n_threads=None,
    progress_bar=True,
    leave=True,
):
    """
    Multi-threaded mapping of a function to an iterable. Useful for I/O bound tasks.

    This function allows you to apply a target function to elements in an iterable concurrently using multiple threads.
    You can control the number of threads, introduce time delays between thread launches, and enable a progress bar.

    Parameters:
        target (Callable): The function to apply to each element in the iterable.
        args_list (Iterable): An iterable of arguments to be passed to the target function in parallel.
        time_delay (float, optional): Time delay (in seconds) between launching threads. Default is 0.02.
        timeout (float, optional): The maximum amount of time (in seconds) to wait for a thread to finish. Default is None, which means wait indefinitely.
        max_n_threads (int, optional): The maximum number of threads to run concurrently. Default is None, which means entire args_list will be processed concurrently.
        progress_bar (bool, optional): Enable or disable the progress bar. Default is True.
        leave (bool, optional): Whether to leave the progress bar after completion. Default is True.

    Example:
        .. code-block:: python
            import requests
            from sign_language_translator.utils import threaded_map

            def get_webpage(url, results: dict):
                if url not in results:
                    results[url] = requests.get(url)

            urls = ["https://example.com", "https://github.com", ...]
            results = {}
            args = [(url, results) for url in urls]

            # process urls concurrently, with a maximum of 2 threads at a time
            threaded_map(get_webpage, args, max_n_threads=2)
    """

    if progress_bar:
        args_list = tqdm(args_list, desc="Launching threads", leave=leave)

    threads = []
    for n, args in enumerate(args_list):
        # launch thread
        thread = threading.Thread(target=target, args=args, name=f"{n}_{args}"[:50])
        thread.start()
        threads.append(thread)
        sleep(time_delay)

        # wait for some threads to finish
        if max_n_threads and len(threads) >= max_n_threads:
            for i, thread in enumerate(threads):
                if isinstance(args_list, tqdm):
                    args_list.set_description(f"Awaiting threads {i/len(threads):.0%}")
                thread.join(timeout=timeout)
                # TODO: progress_callback
            # reset threads list
            threads = []
            if isinstance(args_list, tqdm):
                args_list.set_description("Launching threads")

    # wait for remaining threads to finish
    if progress_bar:
        threads = tqdm(threads, desc="Awaiting threads", leave=leave)
    for thread in threads:
        thread.join(timeout=timeout)
        # TODO: progress_callback



"""This module provides utility functions for the sign language translator package.

Functions:
- download: A function for downloading files from urls.
- tree: A function for printing a directory tree.
- sample_one_index: Select an index based on the given probability distribution.
- search_in_values_to_retrieve_key: search inside every dict value and return the key on match.
- in_jupyter_notebook: Checks if the code is running in a Jupyter notebook.
- is_regex: Checks if the given string is a regex or a regular string.
- linear_interpolation: figure out intermediate values inside an array.
- threaded_map: Multi-threaded mapping of a function to an iterable.
- extract_recursive: Recursively extracts values associated with a specified key from a nested dictionary.

Classes:
- ArrayOps: A class for array operations agnostic to numpy.ndarray and torch.Tensor.
- Archive: A utility class for making, viewing and extracting archive files such as .zip files.
- PrintableEnumMeta: A metaclass for making enum classes printable with the class members.
- ProgressStatusCallback: A class for updating a tqdm progress bar inside a function.
"""

from sign_language_translator.utils.archive import Archive
from sign_language_translator.utils.arrays import ArrayOps, linear_interpolation
from sign_language_translator.utils.download import download
from sign_language_translator.utils.parallel import threaded_map
from sign_language_translator.utils.tree import tree
from sign_language_translator.utils.utils import (
    PrintableEnumMeta,
    ProgressStatusCallback,
    extract_recursive,
    in_jupyter_notebook,
    is_regex,
    sample_one_index,
    search_in_values_to_retrieve_key,
)

__all__ = [
    # functions
    "tree",
    "download",
    "search_in_values_to_retrieve_key",
    "sample_one_index",
    "in_jupyter_notebook",
    "is_regex",
    "linear_interpolation",
    "threaded_map",
    "extract_recursive",
    # classes
    "Archive",
    "ArrayOps",
    "PrintableEnumMeta",
    "ProgressStatusCallback",
]



"""text preprocessing module
"""

import re
from typing import Dict


def replace_words(text: str, word_map: Dict[str, str], word_regex: str = r"\w+") -> str:
    def get_replacement(match: re.Match) -> str:
        word = match.group()
        replacement = word_map.get(word, word)
        return replacement

    text = re.sub(word_regex, get_replacement, text)
    return text


def remove_space_before_punctuation(text: str, punctuation={".", ",", "?", "!"}):
    regex = r"\s+[" + "".join([re.escape(punc) for punc in punctuation]) + r"]"

    def get_replacement(match: re.Match) -> str:
        matched_string: str = match.group(0)
        replacement_string = matched_string.lstrip()

        return replacement_string

    fixed_text = re.sub(regex, get_replacement, text)

    return fixed_text


__all__ = [
    "replace_words",
    "remove_space_before_punctuation",
]



"""utility functions for language objects"""

from __future__ import annotations

from typing import TYPE_CHECKING

from sign_language_translator.config.enums import (
    SignLanguages,
    TextLanguages,
    normalize_short_code,
)
from sign_language_translator.languages.sign import PakistanSignLanguage
from sign_language_translator.languages.text import Hindi, Urdu

if TYPE_CHECKING:
    from enum import Enum

    from sign_language_translator.languages.sign import SignLanguage
    from sign_language_translator.languages.text import TextLanguage

# TODO: AbstractFactory to store str to class mappings
def get_text_language(language_name: str | Enum) -> TextLanguage:
    """
    Retrieves a TextLanguage object based on the provided language name.

    Args:
        language_name (str): The name of the language.

    Returns:
        TextLanguage: An instance of the TextLanguage class corresponding to the provided language name.

    Raises:
        ValueError: If no TextLanguage class is known for the provided language name.
    """

    code_to_class = {
        TextLanguages.URDU.value: Urdu,
        TextLanguages.HINDI.value: Hindi,
    }

    class_ = code_to_class.get(normalize_short_code(language_name), None)
    if class_:
        return class_()  # constructor called

    # Unknown
    raise ValueError(f"no text language class known for '{language_name = }'")


def get_sign_language(language_name: str | Enum) -> SignLanguage:
    """
    Retrieves a SignLanguage object based on the provided language name.

    Args:
        language_name (str): The name of the language.

    Returns:
        SignLanguage: An instance of SignLanguage class corresponding to the provided language name.

    Raises:
        ValueError: If no SignLanguage class is known for the provided language name.
    """

    code_to_class = {
        SignLanguages.PAKISTAN_SIGN_LANGUAGE.value: PakistanSignLanguage,
    }

    class_ = code_to_class.get(normalize_short_code(language_name), None)
    if class_:
        return class_()  # constructor called

    # Unknown
    raise ValueError(f"no sign language class known for '{language_name = }'")


__all__ = [
    "get_text_language",
    "get_sign_language",
]



"""text evaluation metrics
"""

from math import log2
from typing import Any, Iterable, Set


class Perplexity:
    """
    A class for calculating the perplexity of sequences based on token frequencies in a corpus.

    Perplexity of a sequence measures how well a language model assigns probabilities to each token in the sequence.
    A lower perplexity indicates that the language model assigns higher probabilities to the tokens in the sequence,
    and therefore, the sequence is more likely to be generated by the language model.
    Conversely, a higher perplexity indicates that the language model assigns lower probabilities to the tokens in the sequence,
    and therefore, the sequence is less likely to be generated by the language model.

    Args:
        all_tokens (set): A set of all tokens in the corpus.
        regularization_constant (float, optional): initial non-zero frequency of tokens.
    """

    def __init__(self, all_tokens: Set[Any], regularizing_constant=1.0) -> None:
        self.token_to_frequency = {token: regularizing_constant for token in all_tokens}
        self.total_frequency = regularizing_constant * len(self.token_to_frequency)

    def update_frequencies(self, corpus: Iterable[Iterable[Any]]):
        """Update the token frequencies based on the given corpus.

        Args:
            corpus (Iterable[Iterable[Any]]): An iterable containing sequences of tokens.
        """

        for sequence in corpus:
            for token in sequence:
                self.token_to_frequency[token] += 1
                self.total_frequency += 1

    def evaluate(self, sequence: Iterable[Any]) -> float:
        """Calculate the perplexity of a given sequence.

        Args:
            sequence (iterable): The sequence of tokens for which perplexity needs to be calculated.

        Returns:
            float: The perplexity value for the given sequence.
        """

        log_probabilities = [
            log2(self.token_to_frequency[token] / self.total_frequency)
            for token in sequence
        ]
        perplexity = 2 ** (-1 / len(sequence) * sum(log_probabilities))  # type: ignore

        return perplexity


# def cosine_similiarity(text1, text2):
#     # word_embedding = word2embd[word]
#     # similarities   = cosine_similarity( word_embedding.reshape(1,-1),
#     #                                     supported_words_embeddings    ).flatten()
#     # smilarities.argsort()[-top_n:] # get indexes of top 5 values in array
#     # top_n_thresh_indexes = to

#     return 1.0


# def word_error_rate(reference, candidate):
#     # pip install jiwer
#     return 1.0


# def rouge(reference, candidate):
#     # use counters
#     ref = set(reference)
#     return len(set(candidate) & ref) / len(ref)


# def bleu(reference, candidate):
#     # use counters
#     can = set(candidate)
#     return len(set(reference) & can) / len(can)


# def f1_score(reference, candidate):
#     bleu_score = bleu(reference, candidate)
#     rogue_score = rouge(reference, candidate)

#     return 2 * bleu_score * rogue_score / ((bleu_score + rogue_score) or 1)


# def normpdf(x: float, mean: float, std: float):
#     denominator = std * ((2.0 * numpy.pi) ** 0.5)
#     numerator = numpy.exp(-0.5 * ((x - mean) / std) ** 2)
#     return numerator / denominator


__all__ = [
    "Perplexity",
]



"""This module provides an abstract base class for video embedding models that
transform a sequence of video frames into an embedding tensor.

Classes:
    - VideoEmbeddingModel: An abstract base class for video embedding models.
"""

from abc import ABC, abstractmethod
from typing import Iterable, Union

from numpy import uint8
from numpy.typing import NDArray
from torch import Tensor


class VideoEmbeddingModel(ABC):
    """
    Abstract base class for video embedding models.

    This class defines the interface for video embedding models, which transform
    a sequence of video frames into an embedding tensor.

    Attributes:
        None

    Methods:
        embed(frame_sequence, **kwargs): Abstract method to embed a sequence of video frames.

    """

    @abstractmethod
    def embed(
        self, frame_sequence: Iterable[Union[Tensor, NDArray[uint8]]], **kwargs
    ) -> Tensor:
        """Embed a sequence of video frames into an embedding tensor.

        Args:
            frame_sequence (Iterable[Union[Tensor, NDArray[uint8]]]): A sequence of video frames,
                where each frame can be either a Tensor or a numpy array of uint8 values of shape (W, H, C).
            **kwargs: Additional keyword arguments specific to the embedding model.

        Returns:
            Tensor: An embedding tensor representing the sequence of video frames.
        """



"""This module defines the ConcatenativeSynthesis class, which represents a rule based model for translating text to sign language."""

from __future__ import annotations

import random
from enum import Enum
from typing import List, Type

from sign_language_translator.config.enums import SignFormats
from sign_language_translator.languages import get_sign_language, get_text_language
from sign_language_translator.languages.sign import SignLanguage
from sign_language_translator.languages.text import TextLanguage
from sign_language_translator.models.text_to_sign.t2s_model import TextToSignModel
from sign_language_translator.vision._utils import get_sign_wrapper_class
from sign_language_translator.vision.sign.sign import Sign


class ConcatenativeSynthesis(TextToSignModel):
    """A class representing a Rule-Based model for translating text to sign language
    by concatenating sign language videos.
    """

    def __init__(
        self,
        text_language: str | TextLanguage | Enum,
        sign_language: str | SignLanguage | Enum,
        sign_format: str | Type[Sign],
    ) -> None:
        """
        Args:
            text_language (str | TextLanguage | Enum): (source) The text language processor object or its identifier. (e.g. "urdu" or slt.languages.text.Urdu)
            sign_language (str | SignLanguage | Enum): (target) The sign language processor object or its identifier. (e.g. "pk-sl" or slt.languages.sign.PakistanSignLanguage)
            sign_format (str | Type[Sign]): (format) The sign features used for mapping labels to sign features. (e.g. "video" or slt.Video)
        """
        self._text_language = self.__get_text_language_object(text_language)
        self._sign_language = self.__get_sign_language_object(sign_language)
        self._sign_format = self.__get_sign_format_class(sign_format)

    @property
    def text_language(self):
        """An object of TextLanguage class or its child that defines preprocessing, tokenization & other NLP functions."""
        return self._text_language

    @text_language.setter
    def text_language(self, text_language: str | TextLanguage | Enum):
        self._text_language = self.__get_text_language_object(text_language)

    @property
    def sign_language(self):
        """An object of SignLanguage class or its child that defines the mapping rules & grammar."""
        return self._sign_language

    @sign_language.setter
    def sign_language(self, sign_language: str | SignLanguage | Enum):
        self._sign_language = self.__get_sign_language_object(sign_language)

    @property
    def sign_format(self):
        """
        The format of the sign language (e.g. slt.Vision.sign.sign.Sign or subclass).

        Class that wraps the sign language features e.g. videos or images. This class can load the signs from dataset and concatenate its objects
        """

        return self._sign_format

    @sign_format.setter
    def sign_format(self, sign_format: str | Type[Sign]):
        self._sign_format = self.__get_sign_format_class(sign_format)

    def translate(self, text: str, *args, **kwargs) -> Sign:
        """
        Translate text to sign language.

        Args:
            text: The input text to be translated.

        Returns:
            The translated sign language sentence.

        """

        sign_language_sentence = None
        video_labels = []

        text = self.text_language.preprocess(text)
        sentences = self.text_language.sentence_tokenize(text)
        for sentence in sentences:
            tokens = self.text_language.tokenize(sentence)
            tags = self.text_language.get_tags(tokens)

            tokens, tags, contexts = self.sign_language.restructure_sentence(
                tokens, tags=tags
            )
            sign_dicts = self.sign_language.tokens_to_sign_dicts(
                tokens, tags=tags, contexts=contexts
            )

            video_labels.extend(
                [
                    label
                    for sign_dict in sign_dicts
                    for label in random.choices(
                        sign_dict[self.sign_language.SignDictKeys.SIGNS.value],
                        weights=sign_dict[self.sign_language.SignDictKeys.WEIGHTS.value],  # type: ignore
                        k=1,
                    )[0]
                ]
            )

        signs = self._map_labels_to_sign(video_labels)
        # TODO: Trim signs where hand is just being raised from or lowered to the resting position
        sign_language_sentence = self.sign_format.concatenate(signs)

        return sign_language_sentence

    def _map_labels_to_sign(self, labels: List[str]) -> List[Sign]:
        return [
            self.sign_format(self._prepare_resource_name(label)) for label in labels
        ]

    def _prepare_resource_name(self, label, person=None, camera=None, sep="_"):
        if person is not None:
            label = f"{label}{sep}{person}"
        if camera is not None:
            label = f"{label}{sep}{camera}"

        directory = (
            "videos"
            if self.sign_format.name() == SignFormats.VIDEO.value
            else "landmarks"
        )
        name = f"{directory}/{label}.mp4"

        return name

    def __get_text_language_object(
        self, text_language: str | TextLanguage | Enum
    ) -> TextLanguage:
        if isinstance(text_language, (str, Enum)):
            return get_text_language(text_language)
        if isinstance(text_language, TextLanguage):
            return text_language
        raise TypeError(f"Expected str or TextLanguage, got {type(text_language) = }.")

    def __get_sign_language_object(
        self, sign_language: str | SignLanguage | Enum
    ) -> SignLanguage:
        if isinstance(sign_language, (str, Enum)):
            return get_sign_language(sign_language)
        if isinstance(sign_language, SignLanguage):
            return sign_language
        raise TypeError(f"Expected str or SignLanguage, got {type(sign_language) = }.")

    def __get_sign_format_class(self, sign_format: str | Type[Sign]) -> Type[Sign]:
        if isinstance(sign_format, str | Enum):
            return get_sign_wrapper_class(sign_format)
        if isinstance(sign_format, type) and issubclass(sign_format, Sign):
            return sign_format
        raise TypeError(f"Expected str or type[Sign], got {sign_format = }.")



"""
This module contains the `MediaPipeLandmarksModel` class, which is a deep learning-based
video embedding model utilizing the MediaPipe framework for extracting pose and hand
landmarks from video frames.

Classes:
    MediaPipeLandmarksModel: A video embedding model that utilizes MediaPipe for pose and hand landmark extraction.

Example:
    ```
    from sign_language_translator.models import MediaPipeLandmarksModel
    from sign_language_translator.vision.utils import iter_frames_with_opencv

    mediapipe_model = MediaPipeLandmarksModel(number_of_persons=1)

    frame_sequence = iter_frames_with_opencv("video.mp4")
    embedding = mediapipe_model.embed(frame_sequence, landmark_type="world")
    print(embedding.shape)
    ```
"""

from os.path import join
from typing import Dict, Iterable, List

import mediapipe
import numpy as np
import torch
from numpy.typing import NDArray

from sign_language_translator.config.assets import Assets
from sign_language_translator.models.video_embedding.video_embedding_model import (
    VideoEmbeddingModel,
)
from sign_language_translator.utils import ProgressStatusCallback


class MediaPipeLandmarksModel(VideoEmbeddingModel):
    """
    A video embedding model using MediaPipe to extract pose and hand landmarks from video frames.

    Args:
        pose_model_name (str): The name of the pose estimation model.
        hand_model_name (str): The name of the hand estimation model.
        number_of_persons (int): The maximum number of persons to detect in each frame.

    Attributes:
        n_persons (int): The maximum number of persons to detect in each frame.

    Methods:
        embed: Embeds a sequence of frames using pose and hand landmarks.
    """

    def __init__(
        self,
        pose_model_name="pose_landmarker_heavy.task",
        hand_model_name="hand_landmarker.task",
        number_of_persons: int = 1,
    ) -> None:
        self._pose_class = mediapipe.tasks.vision.PoseLandmarker
        self._hand_class = mediapipe.tasks.vision.HandLandmarker

        path = self.__download_and_get_model_path(f"models/mediapipe/{pose_model_name}")
        self._pose_options = mediapipe.tasks.vision.PoseLandmarkerOptions(
            base_options=mediapipe.tasks.BaseOptions(model_asset_path=path),
            running_mode=mediapipe.tasks.vision.RunningMode.VIDEO,
            output_segmentation_masks=False,
            num_poses=number_of_persons,
        )

        path = self.__download_and_get_model_path(f"models/mediapipe/{hand_model_name}")
        self._hand_options = mediapipe.tasks.vision.HandLandmarkerOptions(
            base_options=mediapipe.tasks.BaseOptions(model_asset_path=path),
            running_mode=mediapipe.tasks.vision.RunningMode.VIDEO,
            num_hands=number_of_persons * 2,
        )

        self.n_persons = number_of_persons

    def embed(
        self,
        frame_sequence: Iterable[torch.Tensor | NDArray[np.uint8]],
        landmark_type: str = "world" or "image" or "all",
        progress_callback: None | ProgressStatusCallback = None,
        total_frames: None | int = None,
        **kwargs,
    ) -> torch.Tensor:
        """
        Embed a sequence of frames (video) into a sequence of pose & hand landmarks.

        Args:
            frame_sequence (Iterable[torch.Tensor | NDArray[np.uint8]]): A sequence of video frames as 3D arrays (W, H, c).
            landmark_type (str): The type of landmarks to include in the embedding ("world", "image", "all").

        Returns:
            torch.Tensor: A tensor containing the frame embeddings.
        """

        assert landmark_type in ("world", "image", "all"), "landmark type not supported"

        # TODO: Pose only or hands only
        if hasattr(frame_sequence, "__len__"):
            total_frames = len(frame_sequence)  # type: ignore

        embeddings = []

        with (
            # TODO: create here or in __init__ ??
            self._pose_class.create_from_options(self._pose_options) as pose_landmarker,
            self._hand_class.create_from_options(self._hand_options) as hand_landmarker,
        ):
            for i, frame in enumerate(frame_sequence):
                # convert frame to mediapipe image
                mp_image = mediapipe.Image(
                    image_format=mediapipe.ImageFormat.SRGB,
                    data=np.array(frame),
                )

                # infer through models
                pose_result = pose_landmarker.detect_for_video(mp_image, i)
                hand_result = hand_landmarker.detect_for_video(mp_image, i)

                # create & append the frame embedding
                poses = self._extract_from_pose_results(pose_result)
                hands = self._extract_from_hand_results(hand_result)
                persons = self._arange_pose_and_hands(poses, hands)
                frame_embedding = self._create_frame_embedding(persons, landmark_type)

                embeddings.append(frame_embedding)
                if progress_callback and total_frames:
                    progress_callback(
                        {"file": f"{i / total_frames:.1%}" if total_frames else "?%"}
                    )

        return torch.Tensor(embeddings)

    def _flatten_landmarks(self, landmarks) -> List[float]:
        return [
            value
            for lm in landmarks
            for value in [lm.x, lm.y, lm.z, lm.visibility, lm.presence]
        ]

    def _extract_from_pose_results(self, pose_result) -> Dict[str, List[List[float]]]:
        poses = {"image": [], "world": []}

        for pose_image, pose_world in zip(
            pose_result.pose_landmarks, pose_result.pose_world_landmarks
        ):
            poses["image"].append(self._flatten_landmarks(pose_image))
            poses["world"].append(self._flatten_landmarks(pose_world))

        return poses

    def _extract_from_hand_results(self, hand_result) -> Dict[str, List[List[float]]]:
        hands = {
            "Left_image": [],
            "Left_world": [],
            "Right_image": [],
            "Right_world": [],
        }
        for hnd, image, world in zip(
            hand_result.handedness,
            hand_result.hand_landmarks,
            hand_result.hand_world_landmarks,
        ):
            # flatten & separate
            hands[hnd[0].display_name + "_image"].append(self._flatten_landmarks(image))
            hands[hnd[0].display_name + "_world"].append(self._flatten_landmarks(world))

        return hands

    def _arange_pose_and_hands(
        self,
        poses: Dict[str, List[List[float]]],
        hands: Dict[str, List[List[float]]],
    ) -> Dict[str, List[List[float]]]:
        # TODO: Match left & right hands to poses
        # by using minimum distance between hand image centers
        # np.linalg.norm(pose[left_hand_ids].mean(axis=...), hands.mean(axis=...).T).argmin(axis=...)
        default_hand = [0.0] * 5 * 21
        default_pose = [0.0] * 5 * 33

        for k in poses.keys():
            poses[k] += [default_pose] * (self.n_persons - len(poses[k]))

        for k in hands.keys():
            hands[k] += [default_hand] * (self.n_persons - len(hands[k]))

        return {
            key: [
                poses[key][p] + hands["Left_" + key][p] + hands["Right_" + key][p]
                for p in range(self.n_persons)
            ]  # TODO: order of persons should be the same across frames
            for key in ["image", "world"]
        }

    def _create_frame_embedding(
        self, persons: Dict[str, List[List[float]]], landmark_type: str
    ) -> List[float]:
        embedding = []
        # flatten & concat
        if landmark_type in ("world", "all"):
            embedding.extend([value for person in persons["world"] for value in person])
        if landmark_type in ("image", "all"):
            embedding.extend([value for person in persons["image"] for value in person])

        return embedding

    def __download_and_get_model_path(self, model_local_path: str):
        Assets.download(
            model_local_path,
            progress_bar=True,
            leave=False,
            chunk_size=1048576,
        )
        return join(Assets.ROOT_DIR, model_local_path)



"""
This module provides a SynonymFinder class that can find synonyms of a given text
by utilizing translation and back-translation or similarity in embedding vectors.

Dependencies:
- deep_translator

Classes:
    - SynonymFinder: A class for finding synonyms using translation and similarity methods.
"""

from collections import Counter
from typing import List, Optional

from deep_translator import GoogleTranslator
from deep_translator.exceptions import NotValidPayload
from urllib3.exceptions import MaxRetryError, SSLError

from sign_language_translator.utils.parallel import threaded_map


class SynonymFinder:
    """
    This class provides methods for finding synonyms of a given text using two different approaches:
    1. Translation and back-translation through the 'synonyms_by_translation' method (requires internet).
    2. Embedding-based similarity search through the 'synonyms_by_similarity' method.

    Attributes:
        language (str): The target language for translation. Use 2-letter codes (ISO 639-1).
        translator (GoogleTranslator): The translator object for language translation.
        intermediate_languages (List[str]): List of languages supported by the translator, excluding the current language.
        embedding_model (str): The embedding model for similarity-based synonym finding.

    Methods:
        synonyms_by_translation: Finds synonyms by translating text into an intermediate language and then back-translation.
        synonyms_by_similarity: Finds synonyms based on embedding vector similarity.
        translate: Translates text to the specified target language.

    Example:
        .. code-block:: python
            # Instantiate SynonymFinder with the target language
            synonym_finder = SynonymFinder("en")

            # Find synonyms using translation and back-translation
            text = "happy"
            synonyms = synonym_finder.synonyms_by_translation(text)
            print(f"Synonyms by Translation: {synonyms}")

            # Find synonyms using similarity based on embedding vectors
            text = "joyful"
            synonyms = synonym_finder.synonyms_by_similarity(text)
            print(f"Synonyms by Similarity: {synonyms}")
    """

    def __init__(self, language: str) -> None:
        """
        Initialize a SynonymFinder object.

        Args:
            language (str): The target language for translation based synonyms. Use 2-letter codes (ISO 639-1).
        """
        self.language = language

        self._translator = None
        self._intermediate_languages = None
        self._embedding_model = None

    @property
    def translator(self):
        """
        The GoogleTranslator object with the source language as "auto" and the
        target language as the __init__ argument or according to the current state.
        """
        if self._translator is None:
            self._translator = GoogleTranslator(source="auto", target=self.language)
        return self._translator

    @property
    def intermediate_languages(self) -> List[str]:
        """
        Returns a list of languages supported by the translator, excluding the current language.
        They are used to find synonyms by translation and back-translation. These are 2-letter codes (ISO 639-1).
        """
        if not self._intermediate_languages:
            self._intermediate_languages = [
                lang
                for lang in self.translator.get_supported_languages(as_dict=True).values()  # type: ignore
                if lang != self.language
            ]
        return self._intermediate_languages

    @property
    def embedding_model(self):
        if self._embedding_model is None:
            pass  # "paraphrase-distilroberta-base-v1"
        return self._embedding_model

    def synonyms_by_translation(
        self,
        text: str,
        intermediate_languages: Optional[List[str]] = None,
        time_delay: float = 1e-2,
        timeout: float | None = 10,
        max_n_threads: int = 132,
        lower_case: bool = True,
        progress_bar: bool = True,
        leave: bool = False,
    ) -> List[str]:
        """
        Translates the given text into intermediate languages and performs back-translation to obtain synonyms.
        Translation is done via the internet using web scraping by the deep_translator library.

        Args:
            text (str): The text to be translated.
            intermediate_languages (Optional[List[str]]): List of intermediate languages to translate the text into. Use 2-letter codes (ISO 639-1). If None, all supported languages of the translator will be used. Defaults to None.
            time_delay (float): Time delay between translation requests (in seconds). Defaults to 1e-2.
            timeout (float | None): The maximum amount of time (in seconds) to wait for a thread to finish. None means wait indefinitely. Defaults to 10.
            max_n_threads (int): Maximum number of threads to use for parallel translation. Defaults to 128.
            lower_case (bool): Whether to convert the synonyms to lowercase. Defaults to True.
            progress_bar (bool): Whether to display a progress bar during translation. Defaults to True.
            leave (bool): Whether to leave the progress bar after translation. Defaults to True.

        Returns:
            List[str]: A list of synonyms obtained through back-translation from other languages.
        """

        # setup
        if intermediate_languages is None:
            intermediate_languages = self.intermediate_languages

        def translation_function(text: str, target_lang: str, translations: List[str]):
            try:
                translations.append(self.translate(text, target_lang))
            # catch  and pass
            except Exception:
                pass

        # translation into intermediate languages
        translations = []
        threaded_map(
            translation_function,
            [(text, lang, translations) for lang in intermediate_languages],
            time_delay=time_delay,
            timeout=timeout,
            max_n_threads=max_n_threads,
            progress_bar=progress_bar,
            leave=leave,
            # progress_callback=progress_callback,
        )

        # back-translation into source language
        synonyms = []
        threaded_map(
            translation_function,
            [
                (translation.strip(), self.language, synonyms)
                for translation in set(translations + [text])
                if translation.strip()
            ],
            time_delay=time_delay,
            timeout=timeout,
            max_n_threads=max_n_threads,
            progress_bar=progress_bar,
            leave=leave,
        )

        # preprocess
        if lower_case:
            synonyms = [syn.lower() for syn in synonyms]

        # sort by frequency
        synonyms = [
            txt_
            for txt, _ in Counter(synonyms).most_common()
            if (txt_ := txt.strip()) not in ("",)
        ]

        return synonyms

    def translate(self, text: str, target_language: str) -> str:
        """
        Translates the given text to the specified target language.

        Args:
            text (str): The text to be translated.
            target_language (str): The target language for translation. Use 2-letter codes (ISO 639-1).

        Returns:
            str: The translated text.
        """
        try:
            self.translator.target = target_language
            return self.translator.translate(text)
        except (NotValidPayload, MaxRetryError, SSLError):
            return ""

    def synonyms_by_similarity(self, text: str, threshold=0.8) -> List[str]:
        """Looks into a vector database and returns the closest matches to the input text.

        Args:
            text (str): The input text to find synonyms for.
            threshold (float, optional): Cut off value for similarity between embedding vectors. Words with greater similarity score than this value are returned as synonyms. Defaults to 0.8.

        Returns:
            List[str]: A list of synonyms for the input text.
        """
        # TODO: Implement
        return []



"""
sign_language_translator.models.language_models
==============================================

This module contains various language models used in the sign language translator system.

Language Models:
----------------
- NgramLanguageModel: A language model based on n-grams.
- MixerLM: A language model that combines multiple language models using mixing weights or random selection.
- BeamSampling: A class that performs beam search during text generation.
- LanguageModel: An abstract base class for all language models in this package.
- TransformerLanguageModel: A transformer-based language model.

Usage:
------
To use the language models, import them directly from this module. For example:

```python
from sign_language_translator.models.language_models import NgramLanguageModel, TransformerLanguageModel

# Create and use the NgramLanguageModel
model = NgramLanguageModel(window_size=2, unknown_token="")
model.fit(["hello", "world"])
text = "yel"
next_word, probability = model.next(text)

# Create and use the TransformerLanguageModel
transformer_lm = TransformerLanguageModel.load("tlm.pt")
tokens = ["how", "are", "you"]
next_word, probability = transformer_lm.next(tokens)
```
"""

from sign_language_translator.models.language_models import transformer_language_model
from sign_language_translator.models.language_models.abstract_language_model import (
    LanguageModel,
)
from sign_language_translator.models.language_models.beam_sampling import BeamSampling
from sign_language_translator.models.language_models.mixer import MixerLM
from sign_language_translator.models.language_models.ngram_language_model import (
    NgramLanguageModel,
)
from sign_language_translator.models.language_models.transformer_language_model.model import (
    TransformerLanguageModel,
)

__all__ = [
    "NgramLanguageModel",
    "MixerLM",
    "BeamSampling",
    "LanguageModel",
    "TransformerLanguageModel",
    "transformer_language_model",
]



"""Utility functions and classes for PyTorch models.

This module contains various utility functions and classes to assist with PyTorch models and their training.

Functions:
    top_p_top_k_indexes(probabilities, top_p=None, top_k=None):
        Perform top-p (nucleus) and top-k filtering based on the given probabilities.

    plot_lr_scheduler(*args, lr_scheduler_class=None, lr_scheduler_object=None, initial_lr=1e-3, n_steps=20, parameter_group_number=0, **kwargs):
        Plot the learning rate of a specific parameter group across training steps.

    downwards_wave(n_waves, n_steps_per_wave=9, start=1e-3, end=1e-7, amplitude=0.25):
        Generate a downwards wave pattern with a combination of sine wave and linear function.

    set_layers_trainability_(model, layers_to_unfreeze=None, layers_to_freeze=None):
        Set the trainability of specified layers in the given PyTorch model.

Classes:
    FullyLambdaLR(torch.optim.lr_scheduler.LRScheduler):
        Sets the learning rate of each parameter group to a given function that takes step_num, base_lr & last_lr as args.

    VideoEmbeddingPipeline(slt.models.VideoEmbeddingModel):
        With optional multiprocessing, reads video files from paths, performs forward pass of a model on them and saves the output in specified format.

Example usage:
    See the docstrings of individual functions and classes for usage examples.
"""

from __future__ import annotations

__all__ = [
    "top_p_top_k_indexes",
    "FullyLambdaLR",
    "plot_lr_scheduler",
    "downwards_wave",
    "set_layers_trainability_",
    "VideoEmbeddingPipeline",
]

import multiprocessing
from functools import partial
from glob import glob
from os import makedirs
from os.path import abspath, basename, exists, join
from typing import TYPE_CHECKING, Callable, Dict, Iterable, List, Type
from warnings import warn

import numpy
import torch
from numpy.typing import NDArray
from tqdm.auto import tqdm

from sign_language_translator.vision.utils import iter_frames_with_opencv

if TYPE_CHECKING:
    from sign_language_translator.models.video_embedding import VideoEmbeddingModel


def top_p_top_k_indexes(
    probabilities: Iterable[float],
    top_p: float | None = None,
    top_k: int | None = None,
) -> List[int]:
    """Perform top-p (nucleus) and top-k filtering based on the given probabilities.
    Top-k returns the indices of the top-k elements.
    Top-p returns the indices of the top elements whose sum does not exceed a certain value.

    Args:
        probs (Iterable[float]): A 1-D iterable containing the probabilities of each element. Probabilities must sum to 1.
        top_p (float or None): The threshold probability for top-p sampling (0 to 1). If None, top-p sampling is not applied.
        top_k (int or None): The maximum number of elements to keep for top-k sampling. If None, top-k sampling is not applied.

    Returns:
        torch.Tensor: The indices of the selected elements.

    Example usage:
        >>> selected_indices = top_p_top_k_indexes(
        >>>    probabilities=[0.1, 0.2, 0.15, 0.05, 0.3, 0.2],
        >>>    top_p=0.75,
        >>>    top_k=3,
        >>> )
        [4, 1, 5]

        You can then use the `selected_indices` to gather the actual elements from the original tensor.
        >>> sampled_elements = probs[selected_indices]
        [0.3, 0.2, 0.2]
    """

    if top_p is None and top_k is None:
        return sorted(range(len(probabilities)), key=lambda i: probabilities[i], reverse=True)  # type: ignore

    probs = torch.Tensor(probabilities)

    assert probs.dim() == 1, "The input probabilities iterable must be 1-dimensional."
    assert torch.allclose(  # pylint: disable = no-member
        probs.sum(), torch.Tensor([1])
    ), "probabilities must sum to 1."
    assert top_p is None or 0 <= top_p <= 1, "top_p must be between 0 and 1, or None."
    assert top_k is None or top_k > 0, "top_k must be greater than 0, or None."

    sorted_probs, sorted_indices = torch.sort(  # pylint: disable = no-member
        probs, descending=True
    )

    cumulative_probs = torch.cumsum(sorted_probs, dim=0)  # pylint: disable = no-member

    if top_p is not None:
        # Find indices of the elements whose cumulative probability is below top_p
        sorted_indices_to_remove_mask = cumulative_probs > top_p
        # [False, False, ..., True, True, ...] -> True means the probability too low and is below top_p
        # Shift the indices to the right to keep the first token above the threshold
        sorted_indices_to_remove_mask[1:] = sorted_indices_to_remove_mask[:-1].clone()
        # if the first element itself has probability above top_p, shifting step wouldn't help it from removal.
        sorted_indices_to_remove_mask[0] = False
        # Zero out the probabilities of elements whose index is marked for removal
        sorted_probs[sorted_indices_to_remove_mask] = 0.0

    if top_k is not None:
        # Zero out the probabilities of elements beyond top_k
        sorted_probs[top_k:] = 0.0

    # Get the indices of the selected elements
    selected_indices = sorted_indices[sorted_probs > 0]

    return selected_indices.tolist()


class FullyLambdaLR(torch.optim.lr_scheduler.LRScheduler):
    """Sets the learning rate of each parameter group to a given function
    that takes step_num, base_lr, and last_lr as parameters.
    When last_epoch=-1, sets initial lr as lr.

    Args:
        optimizer (Optimizer): Wrapped optimizer.
        lr_lambda (function): A function which computes the learning rate
            given an integer parameter step_num, initial learning rate and previous learning rate
            for each group in optimizer.param_groups.
        last_epoch (int): The index of last epoch. Default: -1.
        verbose (bool): If ``True``, prints a message to stdout for
            each update. Default: ``False``.

    Example:
        >>> scheduler = FullyLambdaLR(
        >>>    optimizer,
        >>>    lambda step_num, base_lr, last_lr: last_lr * (1.08 if step_num%2 == 0 else 0.8)
        >>> )
        >>> for epoch in range(100):
        >>>     train(...)
        >>>     validate(...)
        >>>     scheduler.step()
        >>>     print(scheduler.get_last_lr()[0])
    """

    def __init__(
        self,
        optimizer,
        lr_lambda: Callable[[int, float, float], float],
        last_epoch=-1,
        verbose=False,
    ):
        self.optimizer = optimizer
        self.lr_lambdas = [lr_lambda] * len(optimizer.param_groups)

        super().__init__(optimizer, last_epoch, verbose)

    def get_lr(self):
        return [
            lmbda(self.last_epoch, base_lr, group["lr"])
            for lmbda, base_lr, group in zip(
                self.lr_lambdas,
                self.base_lrs,
                self.optimizer.param_groups,
            )
        ]


def plot_lr_scheduler(
    *args,
    lr_scheduler_class: Type[torch.optim.lr_scheduler.LRScheduler] | None = None,
    # lr_lambda: Callable[[Any], float] | None = None,
    lr_scheduler_object: torch.optim.lr_scheduler.LRScheduler | None = None,
    initial_lr: float = 1e-3,
    n_steps: int = 20,
    parameter_group_number: int = 0,
    save_fig: bool = False,
    fig_name: str | None = None,
    **kwargs,
):
    """Plot the learning rate of a specific parameter group across training steps.

    This function generates a plot to visualize how the learning rate of a specified
    parameter group changes across training steps. It requires either an existing `lr_scheduler_object`
    or a combination of `lr_scheduler_class`, `args`, and `kwargs` to create a new learning rate scheduler.

    Args:
        lr_scheduler_class (Type[torch.optim.lr_scheduler._LRScheduler], optional):
            The class of the learning rate scheduler. Defaults to None.
        lr_scheduler_object (torch.optim.lr_scheduler._LRScheduler, optional):
            An existing object of a learning rate scheduler class. Defaults to None.
        initial_lr (float, optional): The initial learning rate for the new optimizer object
            needed in case lr_scheduler_object is None. Defaults to 1e-3.
        n_steps (int, optional): The number of epochs/steps to visualize the learning rate changes. Defaults to 20.
        parameter_group_number (int, optional): The index for the optimizer's parameter group to plot the learning rate for. Defaults to 0.
        save_fig (bool, optional): Whether to save the plot instead of showing. Defaults to False.
        fig_name (str, optional): The name of the file to save the plot to. Defaults to None (the class name of the lr_scheduler_class).
        *args: Additional arguments to pass to the lr_scheduler_class when creating a new scheduler.
        **kwargs: Additional keyword arguments to pass to the lr_scheduler_class when creating a new scheduler.

    Example:
        ```
        # Using an existing learning rate scheduler object
        lr_scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=10, gamma=0.1)
        plot_lr_scheduler(lr_scheduler_object=lr_scheduler, n_steps=100)

        # Creating a new learning rate scheduler object
        plot_lr_scheduler(
            lr_scheduler_class=torch.optim.lr_scheduler.ExponentialLR,
            initial_lr=0.01,
            gamma=0.9,
            n_steps=50,
        )
        ```
    """

    import matplotlib.pyplot as plt

    optimizer = None
    if lr_scheduler_object is None:

        class TemporaryModel(torch.nn.Module):
            def __init__(self):
                super(TemporaryModel, self).__init__()
                self.linear = torch.nn.Linear(10, 10)

            def forward(self, x):
                return self.linear(x)

        model = TemporaryModel()
        optimizer = torch.optim.AdamW(model.parameters(), lr=initial_lr)
        if lr_scheduler_class is None:
            lr_scheduler_class = torch.optim.lr_scheduler.LambdaLR
        lr_scheduler_object = lr_scheduler_class(optimizer, *args, **kwargs)  # type: ignore

    x_values = list(range(n_steps))
    y_values = []
    for _ in x_values:
        y_values.append(lr_scheduler_object.get_lr()[parameter_group_number])  # type: ignore
        if optimizer:
            optimizer.step()
        lr_scheduler_object.step()

    plt.plot(
        x_values,
        y_values,
        ".-" if len(x_values) < 50 else "-",
        label=f"lr, param_group_no:{parameter_group_number}",
    )
    plt.xlabel("steps")
    plt.legend()
    if save_fig:
        plt.savefig(fig_name or f"{lr_scheduler_object.__class__.__name__}.png")
    else:
        plt.show()
    plt.close()


def downwards_wave(
    n_waves: int,
    n_steps_per_wave: int = 9,
    start: float = 1e-3,
    end: float = 1e-7,
    amplitude: float = 0.25,
) -> numpy.ndarray:
    """
    Generate a downwards wave pattern with a combination of sine wave and linear function.

    The function generates a sequence of points forming a downward wave pattern, which
    consists of a combination of a sine wave and a linear function. The sine wave is
    modulated by the linear function to create a gradual decrease in axis of the
    waves.

    Parameters:
        n_waves (int): Number of peaks/cycles to generate.
        n_steps_per_wave (int, optional): Number of steps per wave. Default is 9.
        start (float, optional): Starting value for the linear function. Default is 1e-3.
        end (float, optional): Ending value for the linear function. Default is 1e-7.
        amplitude (float, optional): Amplitude of the sine wave. Default is 0.25.

    Returns:
        numpy.ndarray: Array containing the y-axis values of downwards wave pattern.
    """

    x = numpy.linspace(0, n_waves, n_waves * n_steps_per_wave)

    downwards_line = numpy.linspace(start, end, len(x))

    wave = numpy.sin(x * 2 * numpy.pi - numpy.pi / 2) + 1
    wave *= abs(start - end) / n_waves * amplitude

    final = wave + downwards_line
    return final


def set_layers_trainability_(
    model: torch.nn.Module,
    layers_to_unfreeze: List[str] | None = None,
    layers_to_freeze: List[str] | None = None,
):
    """
    Set the trainability of specified layers in the given PyTorch model.

    This function allows you to selectively freeze or unfreeze specific layers of a PyTorch model by setting their
    `requires_grad` attribute accordingly.

    Args:
        model (torch.nn.Module): The PyTorch model whose layers' requires_grad will be modified.
        layers_to_unfreeze (List[str] | None, optional): A list of layer names or prefixes for layers
            that you want to unfreeze. If None, no layers will be unfrozen. If [""], all layers will be unfrozen. Default is None.
        layers_to_freeze (List[str] | None, optional): A list of layer names or prefixes for layers
            that you want to freeze. If None, no layers will be frozen. If [""], all layers will be frozen. Default is None.

    Returns:
        None: This function modifies the model in-place. It does not return anything.

    Note:
        - If both `layers_to_unfreeze` and `layers_to_freeze` are None or empty, no action will be taken,
          and the function will return immediately.
        - The layers' names or prefixes specified in the lists should match the names as returned by
          `model.named_parameters()`.

    Examples:
        1. To freeze all layers in the model:
        set_layers_trainability_(model, layers_to_freeze=[""])

        2. To unfreeze the layers with names starting with 'classifier' and 'fc':
        set_layers_trainability_(model, layers_to_unfreeze=["classifier", "fc"])

        3. To unfreeze all layers:
        set_layers_trainability_(model, layers_to_unfreeze=[""])
    """

    if not (layers_to_unfreeze or layers_to_freeze):
        return

    for layer_name, param in model.named_parameters():
        for given_name in layers_to_freeze or []:
            if layer_name.startswith(given_name):
                param.requires_grad = False
        for given_name in layers_to_unfreeze or []:
            if layer_name.startswith(given_name):
                param.requires_grad = True


class VideoEmbeddingPipeline:
    """
    A class for processing and embedding video data using a slt.models.VideoEmbeddingModel.

    Args:
        model (VideoEmbeddingModel): An instance of the VideoEmbeddingModel class or its child class used for embedding.

    Methods:
        process_video(path, save_format="csv", overwrite=False, output_dir=".", **kwargs):
            Load, embed, and save a video's embedding. kwargs are passed to model.embed().

        process_videos_parallel(path_patterns, n_processes=multiprocessing.cpu_count(),
                                save_format="csv", overwrite=False, output_dir=".", **kwargs):
            Process multiple videos in parallel using multiprocessing. kwargs are passed to model.embed().

    Attributes:
        model (VideoEmbeddingModel): The VideoEmbeddingModel instance used for embedding.
    """

    def __init__(self, model: VideoEmbeddingModel):
        self.model = model

    def process_video(
        self, path, save_format="csv", overwrite=False, output_dir=".", **kwargs
    ):
        """
        Load a video, embed its frames, and save the embedding.

        Args:
            path (str): The path to the video file.
            save_format (str, optional): Format for saving the embedding ("csv", "torch", "npy", "npz").
            overwrite (bool, optional): Whether to overwrite existing embedding files.
            output_dir (str, optional): Directory to save the embedding file.
            **kwargs: Additional keyword arguments for embedding model.

        Returns:
            None
        """

        # TODO: handle batched data

        video = self.__read_video(path)
        embedding = self.__embed_video(video, **kwargs)
        # TODO: frames progress callback
        self.__save_embedding(
            embedding,
            basename(path),
            output_dir=output_dir,
            file_format=save_format,
            overwrite=overwrite,
        )

    def process_videos_parallel(
        self,
        path_patterns: List[str],
        n_processes=multiprocessing.cpu_count(),
        save_format="csv",
        overwrite=False,
        output_dir=".",
        **kwargs,
    ):
        """
        Process multiple videos in parallel using multiprocessing, embedding their frames.

        Args:
            path_patterns (list): List of file path patterns to match videos e.g. ["dataset/*.mp4", "dataset/*.avi"].
            n_processes (int, optional): Number of parallel processes. Defaults to multiprocessing.cpu_count().
            save_format (str, optional): Format for saving the embeddings ("csv", "torch", "npy", "npz").
            overwrite (bool, optional): Whether to overwrite existing embedding files.
            output_dir (str, optional): Directory to save the embedding files.
            **kwargs: Additional keyword arguments for embedding model.

        Returns:
            None
        """

        paths = {abspath(path) for pattern in path_patterns for path in glob(pattern)}

        # warn if multiple paths have the same base name
        base_to_paths: Dict[str, List[str]] = {}
        for path in paths:
            if (base := basename(path)) not in base_to_paths:
                base_to_paths[base] = []
            base_to_paths[base].append(path)

        clashing_paths = [
            path for paths in base_to_paths.values() for path in paths if len(paths) > 1
        ]
        if clashing_paths:
            warn(
                "Found multiple paths with the same base name"
                + f" (overwrite=True will prevent skipping). {clashing_paths = }"
            )

        # optionally skip over existing targets
        if not overwrite:
            existing_targets = [
                (join(output_dir, basename(path)) + f".{save_format}") for path in paths
            ]
            existing_targets = {path for path in existing_targets if exists(path)}
            for path in existing_targets:
                warn(
                    f"Target file already exists at {path}. Use overwrite=True to prevent skipping."
                )
            existing_sources = {
                basename(path)[: -len(save_format) - 1] for path in existing_targets
            }
            paths = {path for path in paths if basename(path) not in existing_sources}

        paths = sorted(paths)
        if len(paths) < 1:
            return

        # process
        n_processes = min(n_processes, len(paths), multiprocessing.cpu_count())
        partial_process_video = partial(
            self.process_video,
            save_format=save_format,
            overwrite=overwrite,
            output_dir=output_dir,
            **kwargs,
        )

        if len(paths) == 1:
            list(tqdm(map(partial_process_video, paths), total=1))
            return

        with multiprocessing.Pool(processes=n_processes) as pool:
            list(tqdm(pool.imap(partial_process_video, paths), total=len(paths)))

    def __read_video(self, path) -> Iterable[NDArray[numpy.uint8]]:
        return iter_frames_with_opencv(path)

    def __embed_video(self, video, **kwargs):
        return self.model.embed(video, **kwargs)

    def __save_embedding(
        self,
        embedding: NDArray | torch.Tensor,
        filename: str,
        output_dir=".",
        file_format="csv",
        overwrite=False,
    ):
        target_path = abspath(join(output_dir, filename) + f".{file_format}")

        makedirs(output_dir, exist_ok=True)
        if exists(target_path) and not overwrite:
            warn(f"File already exists at {target_path}")
            return

        if file_format.lower() == "csv":
            numpy.savetxt(target_path, embedding, delimiter=",")
        elif file_format.lower() in ("torch", "pt"):
            torch.save(embedding, target_path)
        elif file_format.lower() == "npy":
            numpy.save(target_path, embedding)
        elif file_format.lower() == "npz":
            numpy.savez_compressed(target_path, **{filename: embedding})



"""Extra Utility functions models placed here to avoid circular imports.

This module contains various utility functions and classes to assist with models.

Functions:
    get_model(model_code: str, sign_language=None, text_language=None, video_feature_model=None):
        Get the model based on the provided model code and optional parameters.
"""

from __future__ import annotations

__all__ = [
    "get_model",
]

from typing import TYPE_CHECKING

from sign_language_translator.config.assets import Assets
from sign_language_translator.config.enums import (
    ModelCodeGroups,
    ModelCodes,
    normalize_short_code,
)

if TYPE_CHECKING:
    from enum import Enum


def get_model(model_code: str | Enum, *args, **kwargs):
    """
    Get the model based on the provided model code and optional parameters.
    See sign_language_translator.config.enums.ModelCodes
    (or slt.ModelCodes) for a list of supported model codes.

    Args:
        model_code (str): The code representing the desired model.

    Returns:
        Any: The instantiated model object if successful, or None if no model found.

    Raises:
        ValueError: If inappropriate argument values are provided for text_language, sign_language, or video_feature_model.
    """

    model_code = normalize_short_code(model_code)
    if model_code == ModelCodes.CONCATENATIVE_SYNTHESIS.value:
        from sign_language_translator.models import ConcatenativeSynthesis

        # TODO: validate arg types
        return ConcatenativeSynthesis(*args, **kwargs)

    if model_code in ModelCodeGroups.ALL_NGRAM_LANGUAGE_MODELS.value:
        from sign_language_translator.models import NgramLanguageModel

        Assets.download(
            f"models/{model_code}", progress_bar=True, leave=False, chunk_size=1048576
        )
        return NgramLanguageModel.load(Assets.get_path(f"models/{model_code}")[0])

    if model_code in ModelCodeGroups.ALL_MIXER_LANGUAGE_MODELS.value:
        from sign_language_translator.models import MixerLM

        Assets.download(
            f"models/{model_code}", progress_bar=True, leave=False, chunk_size=1048576
        )
        return MixerLM.load(Assets.get_path(f"models/{model_code}")[0])

    if model_code in ModelCodeGroups.ALL_TRANSFORMER_LANGUAGE_MODELS.value:
        from sign_language_translator.models import TransformerLanguageModel

        Assets.download(
            f"models/{model_code}", progress_bar=True, leave=False, chunk_size=1048576
        )
        return TransformerLanguageModel.load(Assets.get_path(f"models/{model_code}")[0])

    if model_code in ModelCodeGroups.ALL_MEDIAPIPE_EMBEDDING_MODELS.value:
        from sign_language_translator.models import MediaPipeLandmarksModel

        parts = model_code.split("-")

        pose_version = int(parts[parts.index("pose") + 1])
        # hand_version = int(parts[parts.index("hand") + 1])
        names = ["lite", "full", "heavy"]

        return MediaPipeLandmarksModel(
            pose_model_name=f"pose_landmarker_{names[pose_version]}.task",
            # hand_model_name=f"hand_landmarker_{names[hand_version]}.task",
            # number_of_persons=1,
        )

    return None



"""This module provides a simple n-gram-based statistical language model implementation.

Classes:
- NgramLanguageModel: A simple n-gram-based statistical language model.
"""

from __future__ import annotations

import json
from collections import Counter
from copy import deepcopy
from os.path import exists
from typing import Any, Dict, Iterable, List, Tuple

from sign_language_translator.models.language_models.abstract_language_model import (
    LanguageModel,
)
from sign_language_translator.text.utils import make_ngrams
from sign_language_translator.utils import sample_one_index


class NgramLanguageModel(LanguageModel):
    """NgramLanguageModel is a statistical language model based on n-grams. It provides functionality for training the model on a given training corpus, generating the next token based on a context, and saving/loading the model.

    Attributes:
    - window_size (int): The size of the context window for predicting the next token.
    - unknown_token (str): The token representation used for unknown or out-of-vocabulary tokens.
    - sampling_temperature (float): A temperature parameter controlling the sampling probabilities during token generation.
    - name (str): The name of the language model object (optional).

    Methods:
    - train(self, training_corpus): Alias for the fit() method. Trains the language model on the given training corpus.
    - fit(self, training_corpus): Trains the language model on the given training corpus.
    - finetune(self, training_corpus, weightage: float): Fine-tunes the language model on an additional training corpus with a specified weightage.
    - next(self, context: Iterable) -> Tuple[Any, float]: Samples the next token from the learned distribution based on the given context.
    - next_all(self, context: Iterable) -> Tuple[List[Any], List[float]]: Returns a list of possible next tokens and their associated probabilities based on the given context.
    - load(model_path: str) -> NgramLanguageModel: Deserializes the model from a JSON file.
    - save(self, model_path: str, indent=None, ensure_ascii=False): Serializes the model to a JSON file.
    - __str__(self) -> str: Returns a string representation of the NgramLanguageModel instance.

    Private Methods:
    - _to_key_datatype(self, item: Iterable) -> Tuple: Converts an iterable item to the appropriate datatype for use as a key in the model dictionary.
    - _count_ngrams(self, training_corpus: List[Iterable], n: int) -> Dict[Tuple, int]: Counts the occurrences of n-grams in the training corpus.
    - _group_by_context(self, counts: Dict[Tuple, int]): Groups the n-grams by context and calculates the weights for each next token.
    - _count_parameters(self): Counts the total number of weights/probabilities in the model.

    """

    def __init__(
        self,
        window_size=1,
        unknown_token="<unk>",
        sampling_temperature=1.0,
        name=None,
    ) -> None:
        super().__init__(unknown_token=unknown_token, name=name)
        self.window_size = window_size
        self.n_parameters = 0
        self.sampling_temperature = sampling_temperature
        self.model: Dict[Tuple, Dict[str, List]] = {}

        self._NEXT_TOKEN = "NEXT_TOKEN"
        self._WEIGHTS = "WEIGHTS"

        self._sep = "|||"

    def train(self, training_corpus):
        """Alias for fit(). Trains the language model on the given training corpus.

        Args:
            training_corpus (Iterable[Iterable]): The training corpus, an iterable of sequences representing the text data.

        Returns:
            None
        """

        self.fit(training_corpus)

    def fit(self, training_corpus) -> None:
        """Trains the language model on the given training corpus.

        Args:
            training_corpus (Iterable[Iterable]): The training corpus, an iterable of sequences representing the text data.

        Returns:
            None
        """
        counts = self._count_ngrams(training_corpus, self.window_size + 1)
        counts = self._group_by_context(counts)

        # normalize frequencies into probabilities
        for context, _ in counts.items():
            total = sum(counts[context][self._WEIGHTS])
            counts[context][self._WEIGHTS] = [
                freq / total for freq in counts[context][self._WEIGHTS]
            ]

        self.model = counts
        self.n_parameters = self._count_parameters()

    def finetune(self, training_corpus, weightage: float) -> None:
        """Fine-tunes the language model on an additional training corpus with a specified weightage.

        Args:
            training_corpus (Iterable[Iterable]): The additional training corpus, an iterable of sequences representing the text data.
            weightage (float): The weightage for the additional training corpus, a value between 0.0 and 1.0 (inclusive).
                A weightage of 0.0 means no impact from the additional corpus,
                while a weightage of 1.0 means the model is completely updated based on the additional corpus.

        Returns:
            None

        Raises:
            AssertionError: If the weightage is outside the valid range [0.0, 1.0].
        """

        assert 0.0 <= weightage <= 1.0, "provide 0.0 <= weightage <= 1.0"

        old_model = deepcopy(self.model)
        self.fit(training_corpus)

        for context in old_model:
            if context not in self.model:
                self.model[context] = old_model[context]
            else:
                old_weights = dict(
                    zip(
                        old_model[context][self._NEXT_TOKEN],
                        old_model[context][self._WEIGHTS],
                    )
                )

                # integrate weights of existing next_tokens
                for i, next_token in enumerate(self.model[context][self._NEXT_TOKEN]):
                    old_w = old_weights.pop(next_token, 0.0)
                    self.model[context][self._WEIGHTS][i] *= weightage
                    self.model[context][self._WEIGHTS][i] += old_w * (1 - weightage)

                # append weights of remaining next_tokens
                for next_token, old_w in old_weights.items():
                    self.model[context][self._NEXT_TOKEN].append(next_token)
                    self.model[context][self._WEIGHTS].append(old_w * (1 - weightage))

        self.n_parameters = self._count_parameters()

    def next(self, context: Iterable) -> Tuple[Any, float]:
        next_tokens, probabilities = self.next_all(context)

        index = sample_one_index(probabilities, temperature=self.sampling_temperature)
        next_token, probability = next_tokens[index], probabilities[index]

        return (next_token, probability)

    def next_all(self, context: Iterable) -> Tuple[List[Any], List[float]]:
        context = self._to_key_datatype(context[len(context) - self.window_size :])  # type: ignore

        if context not in self.model:
            next_tokens, probabilities = [self.unknown_token], [1.0]
        else:
            next_tokens: List = self.model[context][self._NEXT_TOKEN]
            probabilities: List[float] = self.model[context][self._WEIGHTS]

        return next_tokens, probabilities

    def _to_key_datatype(self, item: Iterable) -> Tuple:
        """Converts an iterable item to the appropriate datatype for use as a key in the model dictionary."""

        return tuple(item)

    def _count_ngrams(
        self, training_corpus: Iterable[Iterable], n: int
    ) -> Dict[Tuple, int]:
        """Counts the occurrences of n-grams in the training corpus."""

        all_ngrams = [
            self._to_key_datatype(ngram)
            for sequence in training_corpus
            for ngram in make_ngrams(sequence, n)
        ]
        counts = dict(Counter(all_ngrams))
        return counts

    def _group_by_context(self, counts: Dict[Tuple, int]):
        """collect all the keys that are the same except the ending part."""

        grouped = {}
        for ngram, freq in counts.items():
            context = ngram[:-1]
            next_token = ngram[-1]

            if context not in grouped:
                grouped[context] = {self._NEXT_TOKEN: [], self._WEIGHTS: []}

            grouped[context][self._NEXT_TOKEN].append(next_token)
            grouped[context][self._WEIGHTS].append(freq)

        return grouped

    def _count_parameters(self):
        """Counts the total number of weights/probabilities in the model."""
        return sum(len(v[self._WEIGHTS]) for v in self.model.values())

    @staticmethod
    def load(model_path: str) -> NgramLanguageModel:
        """Deserializes the model (from JSON).

        Args:
            model_path (str): The source file path.

        Returns:
            NgramLanguageModel: The deserialized NgramLanguageModel instance.
        """

        with open(model_path, "r", encoding="utf-8") as f:
            model_data: Dict[str, Any] = json.load(f)

        window_size = int(model_data["window_size"])
        unknown_token = str(model_data["unknown_token"])
        name = str(model_data["name"])

        next_token_key = str(model_data["NEXT_TOKEN"])
        weights_key = str(model_data["WEIGHTS"])
        _sep = str(model_data["sep"])
        sampling_temperature = float(model_data["sampling_temperature"])

        slm = NgramLanguageModel(
            window_size=window_size,
            unknown_token=unknown_token,
            name=name,
            sampling_temperature=sampling_temperature,
        )
        slm._sep = _sep
        slm._NEXT_TOKEN = next_token_key
        slm._WEIGHTS = weights_key

        slm.model = {
            slm._to_key_datatype(context.split(_sep)): next_token_and_weights
            for context, next_token_and_weights in model_data["model"].items()
        }

        return slm

    def save(
        self, model_path: str, indent=None, ensure_ascii=False, overwrite=False
    ) -> None:
        """
        Serializes the model (as JSON).

        Args:
            model_path (str): The target file path. It will silently overwrite if a file already exists at this path.
            indent (Optional[int]): The indentation level for formatting the JSON data (optional).
            ensure_ascii (bool): Controls whether non-ASCII characters are escaped (optional).
            overwrite (bool): If False, raises FileExistsError if the model already exists. Defaults to False.
        """

        if exists(model_path) and not overwrite:
            raise FileExistsError(f"there is already a file at {model_path = }")

        model_data = dict()
        model_data["window_size"] = self.window_size
        model_data["unknown_token"] = self.unknown_token
        model_data["name"] = self.name

        model_data["NEXT_TOKEN"] = self._NEXT_TOKEN
        model_data["WEIGHTS"] = self._WEIGHTS
        model_data["sep"] = self._sep
        model_data["sampling_temperature"] = self.sampling_temperature

        model_data["model"] = {self._sep.join(k): v for k, v in self.model.items()}

        with open(model_path, "w", encoding="utf-8") as f:
            json.dump(model_data, f, indent=indent, ensure_ascii=ensure_ascii)

    def __str__(self) -> str:
        return f"Ngram LM: {super().__str__()}, window={self.window_size}, params={self.n_parameters}"
